'''
* Copyright 2015-2020 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

import time

from .core import *
from .vtkinteractorstyles import CalcamInteractorStyle3D
from .. import render

opacity_params = {'volume':(1,3.45),'lines':(5,2.66),'wall_coverage':(10,1.17)}

def nextcheckstate(checkbox):
    """
    Cycles a checkbox to its next state. This is used to overload the
    usual nextCheckState() method of a tri-state qCheckBox so we get a
    tri-state checkbox which can't be set partially checked by the user.

    Parameters:
        checkbox : A QCheckBox instance on which to act.

    """
    if checkbox.checkState() == qt.Qt.Checked:
        checkbox.setCheckState(qt.Qt.Unchecked)
    else:
        checkbox.setCheckState(qt.Qt.Checked)


# CAD viewer window.
# This allows viewing of the CAD model and overlaying raycasted sight-lines, etc.
class Viewer(CalcamGUIWindow):
 
    def __init__(self, app, parent = None):

        CalcamGUIWindow.init(self,'viewer.ui',app,parent)

        # Start up with no CAD model
        self.cadmodel = None

        # Set up VTK
        self.qvtkwidget_3d = qt.QVTKRenderWindowInteractor(self.vtk_frame)
        self.vtk_frame.layout().addWidget(self.qvtkwidget_3d,0,0,1,2)
        self.interactor3d = CalcamInteractorStyle3D(refresh_callback=self.refresh_3d,viewport_callback=self.update_viewport_info,newpick_callback=self.add_cursor,cursor_move_callback=self.update_cursor_position,resize_callback=self.on_resize)
        self.qvtkwidget_3d.SetInteractorStyle(self.interactor3d)
        self.renderer_3d = vtk.vtkRenderer()
        self.renderer_3d.SetBackground(0, 0, 0)
        self.qvtkwidget_3d.GetRenderWindow().AddRenderer(self.renderer_3d)
        self.camera_3d = self.renderer_3d.GetActiveCamera()


        self.populate_models()
            

        # Callbacks for GUI elements
        self.viewlist.itemSelectionChanged.connect(self.change_cad_view)
        self.camX.valueChanged.connect(self.change_cad_view)
        self.camY.valueChanged.connect(self.change_cad_view)
        self.camZ.valueChanged.connect(self.change_cad_view)
        self.tarX.valueChanged.connect(self.change_cad_view)
        self.tarY.valueChanged.connect(self.change_cad_view)
        self.tarZ.valueChanged.connect(self.change_cad_view)
        self.cam_roll.valueChanged.connect(self.change_cad_view)
        self.camFOV.valueChanged.connect(self.change_cad_view)
        self.sightlines_list.itemChanged.connect(self.update_sightlines)
        self.lines_3d_list.itemChanged.connect(self.update_lines)
        self.load_model_button.clicked.connect(self.load_model)
        self.feature_tree.itemChanged.connect(self.update_checked_features)
        self.feature_tree.itemSelectionChanged.connect(self.update_cadtree_selection)
        self.xsection_checkbox.toggled.connect(self.update_xsection)
        self.xsection_origin.toggled.connect(self.update_xsection)
        self.sightline_opacity_slider.valueChanged.connect(self.update_sightlines)
        self.rendertype_edges.toggled.connect(self.toggle_wireframe)
        self.viewport_load_calib.clicked.connect(self.load_viewport_calib)
        self.sightlines_load_button.clicked.connect(self.update_sightlines)
        self.pick_sightlines_colour.clicked.connect(self.update_sightlines)
        self.sightlines_list.itemSelectionChanged.connect(self.update_selected_sightlines)
        self.sightline_type_volume.clicked.connect(self.update_sightlines)
        self.sightline_type_lines.clicked.connect(self.update_sightlines)
        self.sightline_type_wallcoverage.clicked.connect(self.update_sightlines)
        self.sightlines_legend_checkbox.toggled.connect(self.update_sightlines)
        self.render_button.clicked.connect(self.do_render)
        self.render_cam_view.toggled.connect(self.change_render_type)
        self.render_unfolded_view.toggled.connect(self.change_render_type)
        self.centre_at_cursor.toggled.connect(self.update_unfolded_render_settings)
        self.unfolded_auto_size.toggled.connect(self.update_unfolded_render_settings)
        self.render_coords_combobox.currentIndexChanged.connect(self.update_render_coords)
        self.render_load_button.clicked.connect(self.load_render_result)
        self.cad_colour_reset_button.clicked.connect(self.set_cad_colour)
        self.cad_colour_choose_button.clicked.connect(self.set_cad_colour)
        self.save_view_button.clicked.connect(self.save_view_to_model)
        self.save_colours_button.clicked.connect(self.save_cad_colours)
        self.model_name.currentIndexChanged.connect(self.populate_model_variants)
        self.contour_off.clicked.connect(self.update_contour)
        self.contour_2d.clicked.connect(self.update_contour)
        self.contour_3d.clicked.connect(self.update_contour)
        self.load_lines_button.clicked.connect(self.update_lines)
        self.pick_lines_colour.clicked.connect(self.update_lines)
        self.lines_3d_list.itemSelectionChanged.connect(self.update_selected_lines)
        self.control_sensitivity_slider.valueChanged.connect(lambda x: self.interactor3d.set_control_sensitivity(x*0.01))
        self.rmb_rotate.toggled.connect(self.interactor3d.set_rmb_rotate)
        self.sightlines_legend = None
        self.render_calib = None
        self.enable_lines_checkbox.toggled.connect(self.update_lines)
        self.enable_points_checkbox.toggled.connect(self.update_lines)
        self.line_width_box.valueChanged.connect(self.update_lines)
        self.marker_diameter_box.valueChanged.connect(self.update_lines)
        self.remove_lines_button.clicked.connect(self.update_lines)
        self.coords_legend_checkbox.toggled.connect(self.update_legend)

        self.control_sensitivity_slider.setValue(self.config.mouse_sensitivity)

        self.proj_perspective.toggled.connect(self.set_projection)

        self.model_actors = {}

        self.cursor_angles = None

        self.line_actors = DodgyDict()

        self.contour_actor = None

        self.sightlines = DodgyDict()
        self.colour_q = []
        self.model_custom_colour = None
        self.viewport_calibs = DodgyDict()

        self.colourcycle = ColourCycle()

        self.tabWidget.setTabEnabled(1,False)
        self.tabWidget.setTabEnabled(2,False)
        self.tabWidget.setTabEnabled(3,False)
        self.tabWidget.setTabEnabled(4,False)

        self.render_coords_text.setHidden(True)
        self.render_coords_combobox.setHidden(True)
        self.render_coords_combobox.setHidden(True)
        self.rendersettings_calib_label.setHidden(True)
        self.render_calib_namelabel.setHidden(True)
        self.render_load_button.setHidden(True)
        self.unfolded_render_settings.setHidden(True)

        # Overload the nextCheckState() method of these checkboxes so that I can use them
        # as tri-state from the code but the user can only toggle them between fully checked and
        # un-checked. The "proper" way to do this would be to write my own subclass of qCheckBox
        # which overloads this method and then insert that in the GUI, but since it's built in QT
        # designer I'm doing this instead, which feels a little hacky but works fine :)
        self.enable_lines_checkbox.nextCheckState = lambda : nextcheckstate(self.enable_lines_checkbox)
        self.enable_points_checkbox.nextCheckState = lambda: nextcheckstate(self.enable_points_checkbox)

        # Start the GUI!
        self.show()

        self.interactor3d.init()
        self.views_root_auto.setHidden(False)
        self.qvtkwidget_3d.GetRenderWindow().GetInteractor().Initialize()


    # Load arbitrary 3D lines from file to display
    def update_lines(self,data):

        if self.sender() is self.load_lines_button:

            filename_filter = 'ASCII Data (*.txt *.csv *.dat)'

            filedialog = qt.QFileDialog(self)

            filedialog.setAcceptMode(filedialog.AcceptOpen)
            filedialog.setFileMode(filedialog.ExistingFile)

            filedialog.setWindowTitle('Open...')
            filedialog.setNameFilter(filename_filter)
            filedialog.exec()

            if qt.qt_ver < 6:
                accepted = filedialog.result() == 1
            else:
                accepted = filedialog.result() == filedialog.Accepted

            if accepted:
                fname = str(filedialog.selectedFiles()[0])

                coords = None
                for delimiter in ['\t',' ',',']:
                    try:
                        coords = np.loadtxt(fname,delimiter=delimiter)
                        lines_name = os.path.split(fname)[1].split('.')[0]

                        if len(coords.shape) == 1:
                            coords = coords[np.newaxis,:]
                    except ValueError:
                        continue

                if coords is None or len(coords.shape) != 2 or coords.shape[1] not in [3,6]:
                    raise UserWarning('Could not load coordinates from the file. Please ensure the file is formatted as N rows, 3 or 6 columns and is tab, space or comma delimited.')

                else:
                    coords_dialog = CoordsDialog(self,coords.shape)
                    coords_dialog.exec()
                    if coords_dialog.result() == 1:
                        
                        # If the coordinates are in R,Z,phi, convert them to cartesian.
                        if coords_dialog.line_coords_combobox.currentIndex() == 1:
                            x = coords[:,0] * np.cos(coords[:,2])
                            y = coords[:,0] * np.sin(coords[:,2])
                            coords[:,2] = coords[:,1]
                            coords[:,0] = x
                            coords[:,1] = y

                            if coords.shape[1] == 6:
                                x = coords[:,3] * np.cos(coords[:,5])
                                y = coords[:,3] * np.sin(coords[:,5])
                                coords[:,5] = coords[:,4]
                                coords[:,3] = x
                                coords[:,4] = y


                        # Add it to the lines list
                        listitem = qt.QListWidgetItem(lines_name)

                        self.line_actors[listitem] = render.CoordsActor(coords,coords_dialog.show_lines.isChecked(),coords_dialog.show_points.isChecked())

                        listitem.setFlags(listitem.flags() | qt.Qt.ItemIsEditable | qt.Qt.ItemIsSelectable)
                        listitem.setToolTip(lines_name)
                        self.lines_3d_list.addItem(listitem)
                        listitem.setCheckState(qt.Qt.Checked)
                        self.lines_3d_list.setCurrentItem(listitem)


        elif self.sender() is self.lines_3d_list:

            if data.checkState() == qt.Qt.Checked:
                self.interactor3d.add_extra_actor(self.line_actors[data])
            else:
                self.interactor3d.remove_extra_actor(self.line_actors[data])

            self.refresh_3d()


        elif self.sender() is self.pick_lines_colour and len(self.lines_3d_list.selectedItems()) > 0:

            current_colour = self.line_actors[self.lines_3d_list.selectedItems()[0]].colour
            picked_colour = self.pick_colour(current_colour)

            if picked_colour is not None:
                for item in self.lines_3d_list.selectedItems():
                    self.line_actors[item].set_colour(picked_colour)

        elif self.sender() is self.enable_lines_checkbox:
            for item in self.lines_3d_list.selectedItems():
                self.line_actors[item].set_lines(data)
            if data:
                self.line_width_box.setEnabled(True)
            else:
                self.line_width_box.setEnabled(False)
            self.refresh_3d()

        elif self.sender() is self.enable_points_checkbox:
            for item in self.lines_3d_list.selectedItems():
                self.line_actors[item].set_markers(data)
            if data:
                self.marker_diameter_box.setEnabled(True)
            else:
                self.marker_diameter_box.setEnabled(False)
            self.refresh_3d()

        elif self.sender() is self.line_width_box:
            for item in self.lines_3d_list.selectedItems():
                self.line_actors[item].linewidth = self.line_width_box.value()
                if self.line_actors[item].lines:
                    self.line_actors[item].set_lines(False)
                    self.line_actors[item].set_lines(True)
            self.refresh_3d()

        elif self.sender() is self.marker_diameter_box:
            for item in self.lines_3d_list.selectedItems():
                self.line_actors[item].markersize = self.marker_diameter_box.value() / 100
            if self.line_actors[item].markers:
                self.line_actors[item].set_markers(False)
                self.line_actors[item].set_markers(True)
            self.refresh_3d()

        elif self.sender() is self.remove_lines_button:
            for item in self.lines_3d_list.selectedItems():
                self.interactor3d.remove_extra_actor(self.line_actors[item])
                self.lines_3d_list.takeItem(self.lines_3d_list.row(item))
                del self.line_actors[item]
            self.refresh_3d()

        self.update_legend()

    def get_fov_opacity(self,actor_type):

        slider_value = self.sightline_opacity_slider.value() / 100
        omin,gamma = opacity_params[actor_type]
        delta_tot = 100 - omin

        range_pos = slider_value**gamma

        return (omin + range_pos*delta_tot)/100


    def update_unfolded_render_settings(self,enable):

        if self.sender() is self.centre_at_cursor:
            if enable:
                self.toroidal_centre_box.setEnabled(False)
                self.poloidal_centre_box.setEnabled(False)
                self.toroidal_centre_box.setValue(self.cursor_angles[0])
                self.poloidal_centre_box.setValue(self.cursor_angles[1])
            else:
                self.toroidal_centre_box.setEnabled(True)
                self.poloidal_centre_box.setEnabled(True)

        elif self.sender() is self.unfolded_auto_size:
            if enable:
                self.unfolded_w.setEnabled(False)
            else:
                self.unfolded_w.setEnabled(True)

    def set_fov_opacity(self,opacity,actor_type):

        omin, gamma = opacity_params[actor_type]
        delta_tot = 100 - omin

        opacity = (opacity*100 - omin)/delta_tot
        opacity = opacity**(1/gamma)

        self.sightline_opacity_slider.blockSignals(True)
        self.sightline_opacity_slider.setValue(opacity * 100)
        self.sightline_opacity_slider.blockSignals(False)


    def update_unfolded_render_settings(self,enable):

        if self.sender() is self.centre_at_cursor:
            if enable:
                self.toroidal_centre_box.setEnabled(False)
                self.poloidal_centre_box.setEnabled(False)
                self.toroidal_centre_box.setValue(self.cursor_angles[0])
                self.poloidal_centre_box.setValue(self.cursor_angles[1])
            else:
                self.toroidal_centre_box.setEnabled(True)
                self.poloidal_centre_box.setEnabled(True)

        elif self.sender() is self.unfolded_auto_size:
            if enable:
                self.unfolded_w.setEnabled(False)
            else:
                self.unfolded_w.setEnabled(True)

    def update_selected_lines(self):

        if len(self.lines_3d_list.selectedItems()) > 0:

            lines = 0
            markers = 0
            linewidth = []
            markersize = []
            for item in self.lines_3d_list.selectedItems():
                if self.line_actors[item].lines:
                    lines += 1
                if self.line_actors[item].markers:
                    markers += 1
                linewidth.append(self.line_actors[item].linewidth)
                markersize.append(self.line_actors[item].markersize*100)

            linewidth = int(np.round(np.mean(linewidth)))
            markersize = np.mean(markersize)

            if lines == len(self.lines_3d_list.selectedItems()):

                linecheckstate = qt.Qt.Checked
            elif lines == 0:
                linecheckstate = qt.Qt.Unchecked
            elif lines < len(self.lines_3d_list.selectedItems()):
                linecheckstate = qt.Qt.PartiallyChecked

            if markers == len(self.lines_3d_list.selectedItems()):
                markercheckstate = qt.Qt.Checked
            elif markers == 0:
                markercheckstate = qt.Qt.Unchecked
            elif markers < len(self.lines_3d_list.selectedItems()):
                markercheckstate = qt.Qt.PartiallyChecked

            self.line_width_box.blockSignals(True)
            self.line_width_box.setValue(linewidth)
            self.line_width_box.blockSignals(False)

            self.marker_diameter_box.blockSignals(True)
            self.marker_diameter_box.setValue(markersize)
            self.marker_diameter_box.blockSignals(False)

            self.enable_lines_checkbox.blockSignals(True)
            self.enable_lines_checkbox.setCheckState(linecheckstate)
            self.enable_lines_checkbox.blockSignals(False)

            if linecheckstate == qt.Qt.Unchecked:
                self.line_width_box.setEnabled(False)
            else:
                self.line_width_box.setEnabled(True)

            self.enable_points_checkbox.blockSignals(True)
            self.enable_points_checkbox.setCheckState(markercheckstate)
            self.enable_points_checkbox.blockSignals(False)

            if markercheckstate == qt.Qt.Unchecked:
                self.marker_diameter_box.setEnabled(False)
            else:
                self.marker_diameter_box.setEnabled(True)

            self.lines_appearance_box.setEnabled(True)
            self.remove_lines_button.setEnabled(True)

        else:
            self.lines_appearance_box.setEnabled(False)
            self.remove_lines_button.setEnabled(False)


    def on_close(self):
        self.qvtkwidget_3d.close()


    def set_projection(self):

        if self.proj_perspective.isChecked():
            self.interactor3d.set_projection('perspective')
        elif self.proj_orthographic.isChecked():
            self.interactor3d.set_projection('orthographic')

        self.update_viewport_info()
        self.refresh_3d()


    def on_view_changed(self):

        self.xsection_checkbox.blockSignals(True)
        self.xsection_cursor.blockSignals(True)
        self.xsection_origin.blockSignals(True)

        if self.interactor3d.get_xsection() is not None:
            self.xsection_checkbox.setChecked(True)
            if np.all( np.array(self.interactor3d.get_xsection()) == 0):
                self.xsection_origin.setChecked(True)
            else:
                self.add_cursor(self.interactor3d.get_xsection())
                self.interactor3d.set_cursor_coords(0,self.interactor3d.get_xsection())
                self.xsection_cursor.setChecked(True)
        else:
            self.xsection_checkbox.setChecked(False)

        self.xsection_checkbox.blockSignals(False)
        self.xsection_cursor.blockSignals(False)
        self.xsection_origin.blockSignals(False)

        if self.interactor3d.get_projection() == 'perspective':
            self.proj_perspective.setChecked(True)
        else:
            self.proj_orthographic.setChecked(True)

        

    def on_change_cad_features(self):
        
        for key,item in self.sightlines:

            recheck = False

            if key.checkState() == qt.Qt.Checked:
                recheck = True
                key.setCheckState(qt.Qt.Unchecked)

            item[1] = None
            if recheck:
                key.setCheckState(qt.Qt.Checked)

        self.update_selected_sightlines()

 
    def toggle_wireframe(self,wireframe):
        
        if self.cadmodel is not None:

            self.cadmodel.set_wireframe( wireframe )

            self.refresh_3d()


    def add_cursor(self,coords):

        if self.interactor3d.focus_cursor is None:
            self.interactor3d.add_cursor(coords)
            self.interactor3d.set_cursor_focus(0)
            self.centre_at_cursor.setEnabled(True)
            self.update_cursor_position(0,coords)


    def update_xsection(self):

        if self.xsection_checkbox.isChecked():
            if self.xsection_cursor.isChecked():
                self.interactor3d.set_xsection(self.interactor3d.get_cursor_coords(0))
            else:
                self.interactor3d.set_xsection((0,0,0))
        else:
            self.interactor3d.set_xsection(None)

        self.interactor3d.update_clipping()
        self.refresh_3d()


    def on_model_load(self):

        self.centre_at_cursor.setChecked(False)

        # Enable the other tabs!
        self.tabWidget.setTabEnabled(1,True)
        self.tabWidget.setTabEnabled(2,True)
        self.tabWidget.setTabEnabled(3,True)
        self.tabWidget.setTabEnabled(4,True)

        self.cad_colour_controls.setEnabled(True)
        self.cad_colour_reset_button.setEnabled(False)
        self.cad_colour_choose_button.setEnabled(False)


        # Enable or disable contour controls as appropriate
        contour_exists = self.cadmodel.wall_contour is not None
        self.contour_off.setEnabled(contour_exists)
        self.contour_2d.setEnabled(contour_exists & (self.interactor3d.focus_cursor is not None) )
        self.contour_3d.setEnabled(contour_exists)
        self.render_unfolded_view.setEnabled(contour_exists)
        self.render_unfolded_view_description.setEnabled(contour_exists)
        if contour_exists:
            self.update_contour()
            self.render_unfolded_view.setToolTip('')
            self.render_unfolded_view_description.setToolTip('')
        else:
            if not self.contour_off.isChecked():
                self.contour_off.toggle()
                self.update_contour()
            if self.render_unfolded_view.isChecked():
                self.render_current_view.toggle()
            self.render_unfolded_view.setToolTip('Rendering un-folded first wall view is only available for CAD models with R,Z wall contours defined.')
            self.render_unfolded_view_description.setToolTip('Rendering un-folded first wall view is only available for CAD models with R,Z wall contours defined.')


    def update_selected_sightlines(self):

        self.sightlines_settings_box.setEnabled(False)

        if len(self.sightlines_list.selectedItems()) > 0:

            for sightlines in self.sightlines_list.selectedItems():

                if self.sightlines[sightlines][1] is not None:

                    self.sightlines_settings_box.setEnabled(True)

                    self.sightline_type_volume.blockSignals(True)
                    self.sightline_type_lines.blockSignals(True)
                    self.sightline_type_wallcoverage.blockSignals(True)
                    if self.sightlines[sightlines][2] == 'volume':
                        self.sightline_type_volume.setChecked(True)
                    elif self.sightlines[sightlines][2] == 'lines':
                        self.sightline_type_lines.setChecked(True)
                    elif self.sightlines[sightlines][2] == 'wall_coverage':
                        self.sightline_type_wallcoverage.setChecked(True)
                    self.sightline_type_volume.blockSignals(False)
                    self.sightline_type_lines.blockSignals(False)
                    self.sightline_type_wallcoverage.blockSignals(False)

                    self.set_fov_opacity(self.sightlines[sightlines][1].GetProperty().GetOpacity(),self.sightlines[sightlines][2])

                    break
            


    def update_contour(self):

        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))

        if self.contour_actor is not None:
            self.interactor3d.remove_extra_actor(self.contour_actor)
            self.contour_actor = None

        if self.contour_2d.isChecked():
            cursor_pos = self.interactor3d.get_cursor_coords(0)
            phi = np.arctan2(cursor_pos[1],cursor_pos[0])
            self.contour_actor = render.get_wall_contour_actor(self.cadmodel.wall_contour,'contour',phi)
            self.contour_actor.GetProperty().SetLineWidth(3)
            self.contour_actor.GetProperty().SetColor((1,0,0))
            self.interactor3d.add_extra_actor(self.contour_actor)

        elif self.contour_3d.isChecked():
            self.contour_actor = render.get_wall_contour_actor(self.cadmodel.wall_contour,'surface')
            self.contour_actor.GetProperty().SetColor((1,0,0))
            self.interactor3d.add_extra_actor(self.contour_actor)

        self.refresh_3d()
        self.app.restoreOverrideCursor()




    def save_cad_colours(self):

        cols = self.cadmodel.get_colour()
        self.cadmodel.set_default_colour(cols)


    def change_render_type(self):

        if self.render_cam_view.isChecked():

            # Check if we have a calibration loaded and if not, load one.
            if self.render_calib is None:
                self.render_button.setEnabled(False)

            self.render_coords_text.setHidden(False)
            self.render_coords_combobox.setHidden(False)
            self.render_coords_combobox.setHidden(False)
            self.rendersettings_calib_label.setHidden(False)
            self.render_calib_namelabel.setHidden(False)
            self.render_load_button.setHidden(False)
            self.unfolded_render_settings.setHidden(True)
            self.render_aa.setHidden(False)
            self.render_aa_label.setHidden(False)
            self.render_resolution.setHidden(False)
            self.render_resolution_label.setHidden(False)
            self.update_render_coords()
            self.render_include_cursor.setHidden(False)
            self.render_bg_label.setHidden(False)
            self.render_bg_black.setHidden(False)
            self.render_bg_transparent.setHidden(False)

        elif self.render_current_view.isChecked():
            self.render_coords_text.setHidden(True)
            self.render_coords_combobox.setHidden(True)
            self.render_coords_combobox.setHidden(True)
            self.rendersettings_calib_label.setHidden(True)
            self.render_calib_namelabel.setHidden(True)
            self.render_load_button.setHidden(True)
            self.render_resolution.setCurrentIndex(-1)
            self.render_button.setEnabled(True)
            self.unfolded_render_settings.setHidden(True)
            self.render_aa.setHidden(False)
            self.render_aa_label.setHidden(False)
            self.render_resolution.setHidden(False)
            self.render_resolution_label.setHidden(False)
            self.interactor3d.on_resize()
            self.render_include_cursor.setHidden(False)
            self.render_bg_label.setHidden(False)
            self.render_bg_black.setHidden(False)
            self.render_bg_transparent.setHidden(False)

        elif self.render_unfolded_view.isChecked():

            self.unfolded_render_settings.setHidden(False)

            self.render_coords_text.setHidden(True)
            self.render_coords_combobox.setHidden(True)
            self.render_coords_combobox.setHidden(True)
            self.rendersettings_calib_label.setHidden(True)
            self.render_calib_namelabel.setHidden(True)
            self.render_load_button.setHidden(True)
            self.render_aa.setHidden(True)
            self.render_aa_label.setHidden(True)
            self.render_button.setEnabled(True)
            self.render_resolution.setHidden(True)
            self.render_resolution_label.setHidden(True)
            self.render_include_cursor.setHidden(True)
            self.render_bg_label.setHidden(True)
            self.render_bg_black.setHidden(True)
            self.render_bg_transparent.setHidden(True)



    def load_render_result(self):
        cal = self.object_from_file('calibration')
        if cal is not None:
            self.render_calib = cal
            self.render_calib_namelabel.setText(cal.name)
            self.render_button.setEnabled(True)
            self.update_render_coords()


    def update_render_coords(self):

        self.render_resolution.clear()
        if self.render_calib is not None:
            if self.render_coords_combobox.currentIndex() == 0:
                base_size = self.render_calib.geometry.get_display_shape()            
                self.render_resolution.addItem('{:d} x {:d} (same as camera)'.format(base_size[0],base_size[1]))
                self.render_resolution.addItem('{:d} x {:d}'.format(base_size[0]*2,base_size[1]*2))
                self.render_resolution.addItem('{:d} x {:d}'.format(base_size[0]*4,base_size[1]*4))

            elif self.render_coords_combobox.currentIndex() == 1:
                base_size = self.render_calib.geometry.get_original_shape()
                self.render_resolution.addItem('{:d} x {:d} (same as camera)'.format(base_size[0],base_size[1]))



    def on_resize(self,vtk_size):

        if self.render_current_view.isChecked():
            index = max(self.render_resolution.currentIndex(),0)
            self.render_resolution.clear()
            self.render_resolution.addItem('{:d} x {:d} (Window Size)'.format(vtk_size[0],vtk_size[1]))
            self.render_resolution.addItem('{:d} x {:d}'.format(vtk_size[0]*2,vtk_size[1]*2))
            self.render_resolution.addItem('{:d} x {:d}'.format(vtk_size[0]*4,vtk_size[1]*4))
            self.render_resolution.setCurrentIndex(index)
        self.update_vtk_size(vtk_size)


    def do_render(self):

        filename = self.get_save_filename('image')
        if filename is None:
            return

        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.statusbar.showMessage('Rendering image to {:s} ...'.format(filename))


        # Values of extra AA and oversampling to use
        aa = (self.render_aa.currentIndex() + 1)
        oversampling=2**(self.render_resolution.currentIndex())
        use_transparency = self.render_bg_transparent.isChecked()


        # Render the current view, using the current renderer.
        if self.render_current_view.isChecked():
        
            # Actors which will be temporarily removed from the scene while we render
            # -----------------------------------------------------------------------
            temp_actors = []

            # For some reason having the legend completely ruins things,
            # so we turn it off.
            #if self.interactor3d.legend is not None:
            #    temp_actors.append(self.interactor3d.legend)

            # Get rid of the cursor unless the user said not to.
            if not self.render_include_cursor.isChecked():
                if self.interactor3d.focus_cursor is not None:
                    temp_actors.append( self.interactor3d.cursors[0]['actor'] )

            for actor in temp_actors:
                self.renderer_3d.RemoveActor(actor)
            # -------------------------------------------------------------------------

            # Do the render
            im = render.render_hires(self.renderer_3d,oversampling=oversampling,aa=aa,transparency=use_transparency,legendactor=self.interactor3d.legend)

            # Add back the temporarily removed actors
            for actor in temp_actors:
                self.renderer_3d.AddActor(actor)

        # Render a calibrated camera's point of view
        elif self.render_cam_view.isChecked():

            # For render_cam_view we need to tell it what to include apart from the cad model.
            # So make a list of the other actors we need to add:
            extra_actors = []
            for listitem,sightlines in self.sightlines:
                if listitem.checkState() == qt.Qt.Checked:
                    extra_actors.append(sightlines[1])
            for listitem,lines_3d in self.line_actors:
                if listitem.checkState() == qt.Qt.Checked:
                    extra_actors.append(lines_3d)                

            if self.contour_actor is not None:
                extra_actors.append(self.contour_actor)

            coords = ['Display','Original'][ self.render_coords_combobox.currentIndex()]
            
            im = render.render_cam_view(self.cadmodel,self.render_calib,extra_actors = extra_actors,oversampling=oversampling,aa=aa,transparency=use_transparency,verbose=False,coords=coords)

        elif self.render_unfolded_view.isChecked():

            # Build a dictionary of arguments to pass to the rendering function
            args = {'cadmodel':self.cadmodel}

            args['extra_actors'] = []

            for listitem,stuff in self.sightlines:
                if listitem.checkState() == qt.Qt.Checked and stuff[2]  == 'wall_coverage':
                    args['extra_actors'].append(stuff[1])

            for listitem,actor in self.line_actors:
                if listitem.checkState() == qt.Qt.Checked:
                    args['extra_actors'].append(actor)

            args['theta_start'] = self.poloidal_centre_box.value() - 180
            args['phi_start'] = self.toroidal_centre_box.value() - 180

            if self.unfolded_manual_size.isChecked():
                args['w'] = self.unfolded_w.value()

            render_dialog = RenderUnfoldedDialog(self,args)
            im = render_dialog.render()
            if render_dialog.canceled:
                self.app.restoreOverrideCursor()
                self.statusbar.clearMessage()
                return
            render_dialog.accept()

        # Save the image!
        im[:,:,:3] = im[:,:,2::-1]
        cv2.imwrite(filename,im)

        self.renderer_3d.Render()

        self.app.restoreOverrideCursor()
        self.statusbar.clearMessage()


        


    def update_sightlines(self,data):

        if self.sender() is self.sightlines_load_button:

            cals = self.object_from_file('calibration',multiple=True)

            for cal in cals:

                any_result = False
                for subview in range(cal.n_subviews):
                    if cal.view_models[subview] is not None:
                        any_result = True
                        # Add it to the sight lines list
                        cal_name = cal.name
                        if cal.n_subviews > 1:
                            cal_name = cal_name + ' [{:s}]'.format(cal.subview_names[subview])
                        listitem = qt.QListWidgetItem(cal_name)
                        listitem.setFlags(listitem.flags() | qt.Qt.ItemIsEditable | qt.Qt.ItemIsSelectable)
                        listitem.setToolTip(cal_name)
                        self.sightlines_list.addItem(listitem)
                        if self.sightline_type_volume.isChecked():
                            actor_type = 'volume'
                        elif self.sightline_type_lines.isChecked():
                            actor_type = 'lines'
                        elif self.sightline_type_wallcoverage.isChecked():
                            actor_type = 'wall_coverage'
                        if cal.n_subviews > 1:
                            sv = subview
                        else:
                            sv = None
                        self.sightlines[listitem] = [cal,None,actor_type,sv,next(self.colourcycle)]
                        listitem.setCheckState(qt.Qt.Checked)
                        self.sightlines_list.setCurrentItem(listitem)

                if not any_result:
                    self.show_msgbox('The calibration file {:s} does not contain any calibration results and will not be loaded.'.format(cal.filename))

        elif self.sender() is self.sightlines_list:

            if data.checkState() == qt.Qt.Checked:
                
                n = 0
                for item in self.sightlines.keys():
                    if item.checkState() == qt.Qt.Checked:
                        n = n + 1

                if n > 1:
                    self.sightlines_legend_checkbox.setChecked(True)

                if self.sightlines[data][1] is None:
                    self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
                    self.statusbar.showMessage('Ray casting camera sight lines...')
                    self.app.processEvents()
                    if self.sightlines[data][2] == 'wall_coverage':
                        actor = render.get_wall_coverage_actor(self.sightlines[data][0],self.cadmodel,resolution=256,subview=self.sightlines[data][3])
                    else:
                        actor = render.get_fov_actor(self.cadmodel,self.sightlines[data][0],self.sightlines[data][2],subview=self.sightlines[data][3])
                    self.statusbar.clearMessage()
                    actor.GetProperty().SetColor(self.sightlines[data][4])
                    actor.GetProperty().SetOpacity( self.get_fov_opacity(self.sightlines[data][2]) )
                    self.sightlines[data][1] = actor
                    self.interactor3d.add_extra_actor(actor)
                    self.app.restoreOverrideCursor()

                else:

                    self.interactor3d.add_extra_actor(self.sightlines[data][1])

                self.update_selected_sightlines()

            else:
                self.interactor3d.remove_extra_actor(self.sightlines[data][1])

                n = 0
                for item in self.sightlines.keys():
                    if item.checkState() == qt.Qt.Checked:
                        n = n + 1

                if n < 2:
                    self.sightlines_legend_checkbox.setChecked(False)

        elif self.sender() is self.sightline_opacity_slider:

            for item in self.sightlines_list.selectedItems():
                self.sightlines[item][1].GetProperty().SetOpacity( self.get_fov_opacity(self.sightlines[item][2]) )


        elif self.sender() is self.pick_sightlines_colour and len(self.sightlines_list.selectedItems()) > 0:

            picked_colour = self.pick_colour(self.sightlines[self.sightlines_list.selectedItems()[0]][1].GetProperty().GetColor())
            if picked_colour is not None:
                for item in self.sightlines_list.selectedItems():
                    self.sightlines[item][1].GetProperty().SetColor( picked_colour )

        elif self.sender() in [self.sightline_type_volume,self.sightline_type_lines,self.sightline_type_wallcoverage]:

            for item in self.sightlines_list.selectedItems():
                item.setCheckState(qt.Qt.Unchecked)
                self.sightlines[item][1] = None
                if self.sightline_type_volume.isChecked():
                    self.sightlines[item][2] = 'volume'
                elif self.sightline_type_lines.isChecked():
                    self.sightlines[item][2] = 'lines'
                elif self.sightline_type_wallcoverage.isChecked():
                    self.sightlines[item][2] = 'wall_coverage'

                item.setCheckState(qt.Qt.Checked)

            
        self.update_legend()

        self.refresh_3d()


    def update_legend(self):

        legend_items = []

        if self.sightlines_legend_checkbox.isChecked():

            for item in self.sightlines.keys():
                if item.checkState() == qt.Qt.Checked and self.sightlines[item][1] is not None:
                    legend_items.append((str(item.text()), self.sightlines[item][1].GetProperty().GetColor()))


        if self.coords_legend_checkbox.isChecked():

            for item in self.line_actors.keys():
                if item.checkState() == qt.Qt.Checked:
                    legend_items.append((str(item.text()), self.line_actors[item].colour))

        self.interactor3d.set_legend(legend_items)
        self.refresh_3d()


    def update_cursor_position(self,cursor_id,position):
        info = 'Cursor location: ' + self.cadmodel.format_coord(position).replace('\n',' | ')
        self.statusbar.showMessage(info)

        self.contour_2d.setEnabled(self.cadmodel.wall_contour is not None)
        
        if self.contour_2d.isChecked():
            self.update_contour()

        self.xsection_cursor.setEnabled(True)
        if self.xsection_checkbox.isChecked():
            self.interactor3d.set_xsection(self.interactor3d.get_cursor_coords(0))

        phi = np.arctan2(position[1], position[0])
        if phi < 0.:
            phi = phi + 2 * 3.14159
        phi = phi / 3.14159 * 180

        if self.cadmodel.wall_contour is not None:
            r_min = self.cadmodel.wall_contour[:,0].min()
            r_max = self.cadmodel.wall_contour[:,0].max()
            z_min = self.cadmodel.wall_contour[:,1].min()
            z_max = self.cadmodel.wall_contour[:,1].max()
            r_mid = (r_max + r_min) / 2
            z_mid = (z_max + z_min) / 2

            dr = np.sqrt(position[0]**2 + position[1]**2) - r_mid
            dz = position[2] - z_mid

            theta = 180*np.arctan2(dz,dr)/np.pi

            self.cursor_angles = (phi,theta)

            if self.centre_at_cursor.isChecked():
                self.toroidal_centre_box.setValue(phi)
                self.poloidal_centre_box.setValue(theta)



class CoordsDialog(qt.QDialog):

    def __init__(self, parent,coords_shape):

        # GUI initialisation
        qt.QDialog.__init__(self, parent)
        qt.uic.loadUi(os.path.join(guipath,'qt_designer_files','line_coords.ui'), self)

        self.parent = parent

        if coords_shape[1] == 6:
            self.lines_label.setText('Importing coordinates for {:d} 3D lines from file.'.format(coords_shape[0]))
        elif coords_shape[1] == 3:
            self.lines_label.setText('Importing a sequence of {:d} points from file.'.format(coords_shape[0]))

        if coords_shape[0] == 1:
            self.show_lines.setChecked(False)
            self.show_lines.setEnabled(False)
            self.show_points.setChecked(True)



class RenderUnfoldedDialog(qt.QDialog):

    def __init__(self, parent,render_args):

        # GUI initialisation
        qt.QDialog.__init__(self, parent,qt.Qt.WindowTitleHint | qt.Qt.WindowCloseButtonHint)
        qt.uic.loadUi(os.path.join(guipath,'qt_designer_files','rendering_progress.ui'), self)

        self.parent = parent
        self.canceled = False
        self.args = render_args

        self.args['progress_callback'] = self.update
        self.args['cancel'] = self.cancel_check_callback

        self.time_remaining_text.hide()

        self.previous_remaining_prediction = None

        self.progress_ref = None

    def render(self):
        self.show()
        return render.render_unfolded_wall(**self.args)

    def update(self,progress):

        self.parent.app.processEvents()
        self.repaint()
        try:
            self.progressbar.setValue(progress*100)
            if self.progress_ref is None:
                self.progress_ref = (time.time(),progress)
            else:
                dt = time.time() - self.progress_ref[0]
                dp = progress - self.progress_ref[1]
                t_remaining = (1-progress) * dt/dp

                if self.previous_remaining_prediction is None:
                    self.previous_remaining_prediction = (time.time(),t_remaining)
                    return
                else:
                    if np.abs( t_remaining + (time.time() - self.previous_remaining_prediction[0]) - self.previous_remaining_prediction[1]) / self.previous_remaining_prediction[1] > 0.2:
                        self.progress_ref = (time.time(), progress)
                        self.previous_remaining_prediction = None
                        self.time_remaining_text.hide()
                        return

                if t_remaining < 60:
                    t_string = '< 1 minute'
                elif t_remaining > 60**2:
                    min_remaining = np.round(t_remaining/60)
                    t_string = '{:.0f}h {:.0f}min'.format(np.floor(min_remaining/60),min_remaining % 60)
                else:
                    t_string = '{:.0f} minutes'.format(t_remaining/60)
                self.time_remaining_text.show()
                self.time_remaining_text.setText('Estimated time remaining: {:s}'.format(t_string))

        except TypeError:
            self.label.setText(progress)

    def cancel_check_callback(self):
        return self.canceled

    def reject(self):
        self.canceled = True
        super().reject()



'''
* Copyright 2015-2022 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

'''
Calcam Setup script, uses pip and setuptools.

This uses the standard setuptools install_requires to specify the "easy" dependencies of SciPy and MatPlotLib,
but deals with OpenCV, VTK and PyQt using special custom logic and separate pip processes. This is because
(1) It seems to work across more platforms & environments, and (2) It allows implementing logic of fallback
dependencies, which as far as I can tell are not possible using the usual python packaging methods.

To make sure the logic in this script gets run at install even with pip, I have to deliberately break wheel building.
Frankly, this is all a horrible hack in terms of how python package installation is supposed to work, but after trying things
on several platforms, this actually makes it work smoothest on the widest range of platforms. So I'm prioritising creating an easy
experience for the user over good practise for Python packaging, and I can only hope that one day before pip stops supporting direct
installation altogether, the official tools provide a way to do this in a more "proper" way.
'''

import sys
import os
import subprocess

if 'bdist_wheel' in sys.argv:
    raise Exception("Calcam currently cannot be distributed as wheels due to having important install-time logic in setup.py.")

# Read version from the version file.
with open(os.path.join(os.path.split(os.path.abspath(__file__))[0],'calcam','__version__'),'r') as ver_file:
    version = ver_file.readline().rstrip()

# Read the readme.
with open(os.path.join(os.path.split(os.path.abspath(__file__))[0],'README.md'),'r',encoding='utf-8') as readme_file:
    readme = readme_file.read()

print('\n***************************************************************')
print('                Calcam v{:s} Setup Script'.format(version))
print('***************************************************************\n')


# Check if we have pip and setuptools and if not try to give the user a friendly message.
try:
    import pip
except Exception:
    print('\nThe calcam installation process requires pip, but module pip could not be imported. Please install pip and try again (see https://pip.pypa.io/en/stable/installing/).')
    exit()

try:
    from setuptools import setup,find_packages
except Exception:
    print('\nThe calcam installation process requires Python setuptools, but module setuptools could not be imported. Please install setuptools and try again.')
    exit()


def test_dependency(test_cmd):
    """
    Check if a module is importable, using an external python process.
    The reason for doing this in a separate process is because otherwise
    if we install a module in a separate process then try to import it
    on a subsequent check in this process,there be segfaults. So to avoid
    segfaults, use this.
    """
    try:
        subprocess.check_call([sys.executable, '-c',test_cmd],stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False


def pip_install(pkg_name):
    """
    A small utility function to call pip in a new process to install a given package.
    It seems setuptools cannot be trusted to locate some dependencies, or their correct versions,
    and pip works more reliably. Also I pass through the "--user" option from this setup.py but that's all.
    """
    if '--user' in [arg.lower() for arg in sys.argv]:
        extra_opts = ['--user']
    else:
        extra_opts = []

    try:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install',pkg_name] + extra_opts,stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False



if 'install' in sys.argv or 'develop' in sys.argv:

    print('Dealing with more difficult dependencies\nbefore handing over to setuptools...\n')

    if test_dependency('from cv2 import __version__'):
        print('OpenCV-Python installation found.')
    else:
        print('No OpenCV-Python installation found - trying to install opencv-python-headless using pip...')
        pip_install('opencv-python-headless')
        if test_dependency('from cv2 import __version__'):
           print('   OK!')
        else:
            print('\nCould not install opencv-python-headless using pip. Please install OpenCV for python (cv2) manually then try installing Calcam again.')
            exit()


    # Check if any supported version of PyQt is already present
    pyqt = test_dependency('from PyQt6.QtCore import *')
    if not pyqt:
        pyqt = test_dependency('from PyQt5.QtCore import *')
    if not pyqt:
        pyqt = test_dependency('from PyQt4.QtCore import *')

    if pyqt:
        print('Compatible PyQt installation found.')
    else:
        print('No compatible PyQt installation found.')

    # If not, try to install one with pip. Any will do.
    if not pyqt:
        print('   Trying to install PyQt6 using pip...')
        pip_install('PyQt6')
        pyqt = test_dependency('from PyQt6.QtCore import *')
        if pyqt: print('      OK!')
    if not pyqt:
        print('      Failed.\n   Trying PyQt5 instead...')
        pip_install('PyQt5')
        pyqt = test_dependency('from PyQt5.QtCore import *')
        if pyqt: print('      OK!')
    if not pyqt:
        print('      Failed.\n   Trying PyQt4 instead...')
        pip_install('PyQt4')
        pyqt = test_dependency('from PyQt4.QtCore import *')
        if pyqt:
            print('      OK!')
        else:
            print('      Failed.')

    # VTK
    if test_dependency('from vtk import vtkVersion;ver=vtkVersion();v=ver.GetVTKMajorVersion()*100+ver.GetVTKMinorVersion();assert v>600 & v<901'):
        print('Compatible VTK installation found.')
        vtk = True
    else:
        print('No compatible VTK installation found - trying to install with pip...')
        pip_install('vtk>=6,<9.1')
        if test_dependency('from vtk import vtkVersion;ver=vtkVersion();v=ver.GetVTKMajorVersion()*100+ver.GetVTKMinorVersion();assert v>600 & v<901'):
            print('   OK!\n')
            vtk = True
        else:
            print('   Failed.\n')
            vtk = False

    # Make sure a file exists at calcam/gui/__executable_path__ before calling setup() - this makes sure it gets marked as part of the package
    # so allows clean uninnstallation, but without having to keep a blank copy of the file in the git repo which causes git confusion.
    open(os.path.join(os.path.split(__file__)[0],'calcam','gui','__executable_path__'), 'a').close()

    print('\nNow handing over to setuptools...\n')

# Actually do the requested setup actions
s = setup(
        name='calcam',
        version=version,
        license='European Union Public License 1.1',
        author='Scott Silburn et. al.',
        author_email='scott.silburn@ukaea.uk',
        description='Spatial calibration tools for science & engineering camera systems.',
        long_description=readme,
        long_description_content_type='text/markdown',
        classifiers=[
                    "Development Status :: 5 - Production/Stable",
                    "Operating System :: Microsoft :: Windows",
                    "Operating System :: MacOS :: MacOS X",
                    "Operating System :: POSIX :: Linux",
                    "Environment :: Console",
                    "Environment :: X11 Applications :: Qt",
                    "Intended Audience :: Science/Research",
                    "Intended Audience :: Developers",
                    "License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)",
                    "Natural Language :: English",
                    "Programming Language :: Python :: 3",
                    "Topic :: Scientific/Engineering :: Physics",
                    "Topic :: Scientific/Engineering :: Visualization",
                    "Topic :: Scientific/Engineering :: Image Processing"
                    ],
        project_urls={
                    'Documentation': 'https://euratom-software.github.io/calcam',
                    'Source': 'https://github.com/euratom-software/calcam/',
                    'Issue Tracker': 'https://github.com/euratom-software/calcam/issues',
                    'Zenodo':'https://doi.org/10.5281/zenodo.1478554'
                    },
        license_files = ('LICENCE.txt',),
        packages=find_packages(),
        package_data={'calcam':['gui/icons/*','gui/qt_designer_files/*.ui','gui/logo.png','builtin_image_sources/*.py','__version__','gui/__executable_path__']},
        entry_points={ 'gui_scripts': [ 'calcam = calcam:start_gui'] },
        zip_safe=False,
        install_requires=['scipy','matplotlib']
        )


if 'install' in sys.argv or 'develop' in sys.argv:
    # Find out where setup() put the the GUI launcher script / executable and do 2 things with it:
    # (1) Print a message telling the user about it,
    # (2) Write it down in a file in the code path so that the user can easily look it up later.
    if 'install' in sys.argv:
        script_dir = s.command_obj['install'].install_scripts
        code_dir = s.command_obj['install'].install_lib
    elif 'develop' in sys.argv:
        script_dir = s.command_obj['develop'].script_dir
        code_dir = os.path.split(__file__)[0]

    executable_path = os.path.join(script_dir.replace('/',os.path.sep),'calcam')
    if sys.platform == 'win32':
        executable_path = executable_path + '.exe'

    extra_msg = '\n\nThe GUI can be launched using the executable:\n{:s}'.format(executable_path)

    try:
        with open(os.path.join(code_dir,'calcam','gui','__executable_path__'),'w') as f:
            f.write(executable_path)
    except Exception:
        pass

    # Check if the executable path is included in the PATH environment variable which makes the GUI easier to run
    try:
        env_path = os.environ['PATH']
    except KeyError:
        env_path = ''

    if script_dir in env_path:
        extra_msg = extra_msg + '\nor just by typing "calcam" at a terminal.'

    print('\n***************************************************************\n\nCalcam installation complete.\n\nIt can be imported as a Python module with "import calcam"{:s}\n\n***************************************************************\n'.format(extra_msg))

else:
    print('\n***************************************************************\n')


'''
* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

""" 
Ray tracing tools for CalCam_py

Written by Scott Silburn
2015-05-17
"""

import time
import os
import random
import copy

try:
    import vtk
except:
    vtk = None

import numpy as np
from scipy.io.netcdf import netcdf_file

from . import coordtransformer
from . import misc
from . import __version__ as calcam_version


def raycast_sightlines(calibration,cadmodel,x=None,y=None,exclusion_radius=0.0,binning=1,coords='Display',verbose=True,intersecting_only=False, force_subview=None,status_callback=None,calc_normals=False):
    '''
    Ray cast camera sight-lines to determine where they intersect the given CAD model.

    Parameters:

        calibration (calcam.Calibration) : Calibration whose sight-lines to raycast.
        
        cadmodel (calcam.CADModel)       : CAD model to check intersection with.
        
        x, y (array-like)                : x and y image pixel coordinates for which to cast sight-lines. \
                                           If not specified, one ray is cast at the centre of every detector pixel.\
                                           x and y must be the same shape.

        exclusion_radius (float)         : Distance from camera pupil (in meters) over which to ignore ray \
                                           intersections with CAD surfaces. \
                                           This is useful for views involving mirrors and/or prisms where unfolding \
                                           the optical path results in the virtual pupil location falling behind \
                                           nearby CAD surfaces. Setting a sufficient exclusion radius will cause the \
                                           rays to be launched from the other side of these surfaces so they intersect \
                                           the correct surfaces present in the image.

        binning (int)                    : If not explicitly providing x and y image coordinates, pixel binning for ray casting.\
                                           This specifies NxN binning, i.e. for a value of 2, one ray is cast at the centre of \
                                           every 2x2 cluster of pixels.
                                           
        coords (str)                     : Either ``Display`` or ``Original``. If specifying x and y coordinates,\
                                           specifies whether the input x and y are in original or display coords. \
                                           Otherwise, specifies the orientation of the returned data.

        intersecting_only (bool)         : If set to True, the ray end coordinates and length for sight-lines which do not intersect \
                                           with the CAD model (i.e. sight lines which "escape" through holes in the model) are set to NaN. \
                                           This is useful if you are only interested in sight-lines which intersect with CAD model surfaces. 
        
        force_subview (int)              : If specified, forces use of the camera model from this index of sub-view \
                                           in the calibration. Otherwise, sub-views are chosen according to the \
                                           sub-view mask in the calibration.
                                           
        verbose (bool)                   : Whether to print status updates during ray casting (depreciated in favour of status_callback)

        status_callback (callable)       : Callable which takes a single argument to be called with status updates. The argument will \
                                           either be a string for textual status updates or a float from 0 to 1 specifying the progress \
                                           of the calculation. If set to None, no status updates are issued. For backwards compatibility, \
                                           if set to None but verbose is set to True, status_callback will be set such that status updates \
                                           go to stdout.

        calc_normals (bool)              : Whether to calculate the normal vectors of the CAD model where the sight-lines intersect it. \
                                           Not turned on by default because it seems to add around 80% extra calculation time, so best used \
                                           only if actyally needed.

    Returns:

        calcam.RayData                   : Object containing the results.

    '''

    if vtk is None:
        raise Exception('VTK is not available, and this Calcam feature requires VTK!')

    if status_callback is None:
        if verbose:
            status_callback = misc.LoopProgPrinter().update
            
    if status_callback is not None:
        original_callback = cadmodel.get_status_callback()
        cadmodel.set_status_callback(status_callback)


    # Work out how big the model is. This is to make sure the rays we cast aren't too short.
    model_extent = cadmodel.get_extent()
    model_size = model_extent[1::2] - model_extent[::2]
    max_ray_length = model_size.max() * 4
    
    if status_callback is not None:
        status_callback('Getting CAD model octree...')

    # Get the CAD model's octree
    cell_locator = cadmodel.get_cell_locator()

    # If no pixels are specified, do the whole chip at the specified binning level.
    fullchip = False
    if x is None and y is None:
        fullchip = True
        x,y = calibration.fullframe_meshgrid(coords=coords,binning=binning)

    elif x is None or y is None:
        raise ValueError('Either both or none of x and y pixel coordinates must be given!')


    if np.array(x).ndim == 0:
        x = np.array([x])
    else:
        x = np.array(x)
    if np.array(y).ndim == 0:
        y = np.array([y])
    else:
        y = np.array(y)
    if x.shape != y.shape:
        raise ValueError('x and y arrays must be the same shape!')

    valid_mask = np.logical_and(np.isnan(x) == 0 , np.isnan(y) == 0 )
    if coords.lower() == 'original':
        x,y = calibration.geometry.original_to_display_coords(x,y)

    results = RayData()
    
    if fullchip:
        results.fullchip = coords
    else:
        results.fullchip = False
        
    results.x = np.copy(x).astype('float')
    results.x[valid_mask == 0] = 0
    results.y = np.copy(y).astype('float')
    results.y[valid_mask == 0] = 0
    results.transform = calibration.geometry
    if calibration.filename is not None:
        splitname = os.path.split(calibration.filename)
        results.history = 'Ray cast of calibration "{:s}" [from: {:s}] by {:s} on {:s} at {:s}'.format(splitname[1].replace('.ccc',''),splitname[0],misc.username,misc.hostname,misc.get_formatted_time())
    else:
        results.history = 'Ray cast by {:s} on {:s} at {:s}'.format(misc.username,misc.hostname,misc.get_formatted_time())

    orig_shape = np.shape(results.x)
    results.x = np.reshape(results.x,np.size(results.x),order='F')
    results.y = np.reshape(results.y,np.size(results.y),order='F')
    valid_mask = np.reshape(valid_mask,np.size(valid_mask),order='F')


    # New results object to store results
    if fullchip:
        results.binning = binning
        results.coords = coords
    else:
        results.binning = None
        results.coords = None


    results.ray_end_coords = np.ndarray([np.size(x),3])
    results.model_normals = np.ndarray([np.size(x),3]) + np.nan

    # Line of sight directions
    LOSDir = calibration.get_los_direction(results.x,results.y,coords='Display',subview=force_subview)
    if len(LOSDir.shape) == 1:
        LOSDir = [LOSDir]
    results.ray_start_coords = calibration.get_pupilpos(results.x,results.y,coords='Display',subview=force_subview)

    
    if status_callback is not None:
        oom = np.floor( np.log(np.size(x)) / np.log(10) / 3. )
        status_callback('Casting {:s} rays...'.format( ['{:.0f}','{:.1f}k','{:.2f}M'][int(oom)].format(np.size(x)/10**(3*oom)) ) )


    # Some variables to give to VTK becasue of its annoying C-like interface
    t = vtk.mutable(0)
    pos = np.zeros(3)
    coords_ = np.zeros(3)
    subid = vtk.mutable(0)
    cellid = vtk.mutable(0)
    cell = vtk.vtkGenericCell()

    last_status_update = 0.
    n_done = 0
    
    # We will do the ray casting in a random order,
    # purely to get better time remaining estimation.
    inds = list(range(np.size(x)))
    random.shuffle(inds)
    
    for ind in inds:

        if not valid_mask[ind]:
            results.ray_end_coords[ind,:] = np.nan
            results.ray_start_coords[ind,:] = np.nan
            continue

        # Do the raycast and put the result in the output array
        raystart = results.ray_start_coords[ind] + exclusion_radius * LOSDir[ind]
        rayend = results.ray_start_coords[ind] + max_ray_length * LOSDir[ind]

        retval = cell_locator.IntersectWithLine(raystart,rayend,1.e-6,t,pos,coords_,subid,cellid,cell)

        if abs(retval) > 0:
            results.ray_end_coords[ind,:] = pos[:]
            if calc_normals:
                v0 = np.array(cell.GetPoints().GetPoint(2)) - np.array(cell.GetPoints().GetPoint(0))
                v1 = np.array(cell.GetPoints().GetPoint(2)) - np.array(cell.GetPoints().GetPoint(1))
                n = np.cross(v0,v1)
                n = n / np.sqrt(np.sum(n**2))
                if np.dot(LOSDir[ind],n) > 0:
                    n = -n
                results.model_normals[ind,:] = n

        elif intersecting_only:
            results.ray_end_coords[ind,:] = np.nan
        else:
            results.ray_end_coords[ind,:] = rayend

        n_done = n_done + 1
        
        if time.time() - last_status_update > 1 and status_callback is not None:
            status_callback(n_done / len(inds))
            last_status_update = time.time()

    if status_callback is not None:
        status_callback(1.)

    results.x[valid_mask == 0] = np.nan
    results.y[valid_mask == 0] = np.nan

    results.ray_end_coords = np.reshape(results.ray_end_coords,orig_shape + (3,),order='F')
    results.ray_start_coords = np.reshape(results.ray_start_coords,orig_shape + (3,),order='F')
    if calc_normals:
        results.model_normals = np.reshape(results.model_normals, orig_shape + (3,), order='F')
    else:
        results.model_normals = None

    results.x = np.reshape(results.x,orig_shape,order='F')
    results.y = np.reshape(results.y,orig_shape,order='F')

    if status_callback is not None:
        cadmodel.set_status_callback(original_callback)

    return results





class RayData:
    '''
    Class representing ray casting results.

    Objects of this class are returned by :func:`calcam.raycast_sightlines`.
    It can also be used to save and load ray cast results to disk.

    Parameters:
        
        filename (str)  : File name of netCDF file containing saved RayData to load. \
                          If not given, an empty RayData object is created.

    '''
    def __init__(self,filename=None):

        self.ray_end_coords = None
        self.ray_start_coords = None

        self.history = None
        '''
        str: Human readable description of where the raydata came from.
        '''

        self.binning = None
        self.transform = None
        self.fullchip = None
        self.x = None
        self.y = None
        self.filename = None
        self.crop = None
        self.model_normals = None
        
        if filename is not None:
            self._load(filename)


    # Save to a netCDF file
    def save(self,filename):
        '''
        Save the RayData to a netCDF file.

        Parameters:

            filename (str) : File name to save to.
        '''
        if not filename.endswith('.nc'):
            filename = filename + '.nc'

        current_crop = self.crop
        self.set_detector_window(None)

        f = netcdf_file(filename,'w')
        f.title = 'Calcam v{:s} RayData (ray cast results) file.'.format(calcam_version)
        f.history = self.history
        f.image_transform_actions = "['" + "','".join(self.transform.transform_actions) + "']"
        f.fullchip = self.fullchip

        f.createDimension('pointdim',3)

        if len(self.x.shape) == 2:
            f.createDimension('udim',self.x.shape[1])
            f.createDimension('vdim',self.x.shape[0])
            rayhit = f.createVariable('RayEndCoords','f4',('vdim','udim','pointdim'))
            raystart = f.createVariable('RayStartCoords','f4',('vdim','udim','pointdim'))
            x = f.createVariable('PixelXLocation','i4',('vdim','udim'))
            y = f.createVariable('PixelYLocation','i4',('vdim','udim'))
            if self.model_normals is not None:
                normals = f.createVariable('ModelNormals', 'f4', ('vdim', 'udim', 'pointdim'))
            
            rayhit[:,:,:] = self.ray_end_coords
            raystart[:,:,:] = self.ray_start_coords
            x[:,:] = self.x
            y[:,:] = self.y
            if self.model_normals is not None:
                normals[:,:,:] = self.model_normals

        elif len(self.x.shape) == 1:
            f.createDimension('udim',self.x.size)
            rayhit = f.createVariable('RayEndCoords','f4',('udim','pointdim'))
            raystart = f.createVariable('RayStartCoords','f4',('udim','pointdim'))
            if self.model_normals is not None:
                normals = f.createVariable('ModelNormals','f4',('udim','pointdim'))

            x = f.createVariable('PixelXLocation','i4',('udim',))
            y = f.createVariable('PixelYLocation','i4',('udim',))

            rayhit[:,:] = self.ray_end_coords
            raystart[:,:] = self.ray_start_coords
            x[:] = self.x
            y[:] = self.y
            if self.model_normals is not None:
                normals[:,:] = self.model_normals
        else:
            raise Exception('Cannot save RayData with >2D x and y arrays!')

        binning = f.createVariable('Binning','i4',())

        if self.binning is not None:
            binning.assignValue(self.binning)
        else:
            binning.assignValue(0)

        f.createDimension('pixelsdim',2)

        xpx = f.createVariable('image_original_shape','i4',('pixelsdim',))
        xpx[:] = [self.transform.x_pixels,self.transform.y_pixels]

        offset = f.createVariable('image_offset','i4',('pixelsdim',))
        offset[:] = self.transform.offset[:]

        pixelaspect = f.createVariable('image_original_pixel_aspect','f4',())
        pixelaspect.assignValue(self.transform.pixel_aspectratio)

        binning.units = 'pixels'
        raystart.units = 'm'
        rayhit.units = 'm'
        x.units = 'pixels'
        y.units = 'pixels'
        f.close()

        self.set_detector_window(current_crop)



    def _load(self,filename):
        '''
        Load RayData from a file.

        Parameters:

            filename (str) : File name to load from.
        '''
        f = netcdf_file(filename, 'r',mmap=False)
        self.filename = filename
               
        self.ray_end_coords = f.variables['RayEndCoords'].data.astype(np.float32)
        self.ray_start_coords = f.variables['RayStartCoords'].data.astype(np.float32)
        self.binning = f.variables['Binning'].data[()]

        self.transform = coordtransformer.CoordTransformer()
        self.transform.set_transform_actions(eval(f.image_transform_actions))
        self.transform.x_pixels = f.variables['image_original_shape'][0]
        self.transform.y_pixels = f.variables['image_original_shape'][1]
        self.transform.pixel_aspectratio = f.variables['image_original_pixel_aspect'].data[()]

        try:
            self.transform.offset = f.variables['image_offset'][:]
        except KeyError:
            pass

        try:
            self.model_normals = f.variables['ModelNormals'].data.astype(np.float32)
        except KeyError:
            self.model_normals = None

        try:
            self.history = f.history.decode('utf-8')
            self.fullchip = f.fullchip
            if not self.fullchip:
                self.binning = None
            elif self.fullchip != True:
                self.fullchip = self.fullchip.decode('utf-8')
        except AttributeError:
            self.history = 'Loaded from legacy raydata file "{:s} by {:s} on {:s} at {:s}'.format(os.path.split(filename)[-1],misc.username,misc.hostname,misc.get_formatted_time())
            if self.binning == 0:
                self.binning = None
                self.fullchip = False
            else:
                if len(self.transform.transform_actions) == 0 and self.transform.pixel_aspectratio == 1:
                    self.fullchip = 'Original'
                else:                
                    self.fullchip = True

        self.x = f.variables['PixelXLocation'].data
        self.y = f.variables['PixelYLocation'].data

        f.close()


    def set_detector_window(self,window):
        '''
        Adjust the raydata to apply to a different detector region for than was used
        to perform the original raycast. Useful for example if a CMOS camera has been calibrated
        over the full frame, but you now want to use this calibration for data which
        has been cropped.

        Calling this function with `None` as the single argument sets the raydata
        back to its "native" state. Otherwise, call with a 4 element tuple specifying the
        left,top,width and height of the detector window.

        Detector window coordinates must always be in original coordinates.

        Parameters:

            window (tuple or list) : A 4-element tuple or list of integers defining the \
                                     detector window coordinates (Left,Top,Width,Height)
        '''
        if window is None:

            if self.crop is not None:

                dx = self.crop[0] - self.native_geometry[2][0]
                dy = self.crop[1] - self.native_geometry[2][1]

                ox, oy = self.transform.display_to_original_coords(self.x, self.y)
                ox = ox + dx
                oy = oy + dy

                self.transform.x_pixels,self.transform.y_pixels,self.transform.offset = self.native_geometry

                self.x,self.y = self.transform.original_to_display_coords(ox,oy)

                del self.native_geometry

                self.crop_inds = None
                self.crop = None

        elif len(window) == 4:

            dx = window[0] - self.transform.offset[0]
            dy = window[1] - self.transform.offset[1]

            ox,oy = self.transform.display_to_original_coords(self.x,self.y)
            ox = ox - dx
            oy = oy - dy

            self.native_geometry = (self.transform.x_pixels,self.transform.y_pixels,self.transform.offset)

            self.transform.x_pixels = window[2]
            self.transform.y_pixels = window[3]
            self.transform.offset = (window[0],window[1])

            nx,ny = self.transform.original_to_display_coords(ox,oy)

            newshape = self.transform.get_display_shape()

            if self.fullchip and (np.all( nx > 0) or np.all(nx + 1 < newshape[0]) or np.all(ny > 0) or np.all(ny + 1 < newshape[1])):
                self.transform.x_pixels,self.transform.y_pixels,self.transform.offset = self.native_geometry

                raise ValueError('Requested crop window ({:d}x{:d} at {:d}x{:d}) is outside the raycasted area ({:d}x{:d} at {:d}x{:d})'.format(window[2],window[3],window[0],window[1],self.transform.x_pixels,self.transform.y_pixels,self.transform.offset[0],self.transform.offset[1]))

            self.x = nx
            self.y = ny

            if self.fullchip:
                rowinds = np.squeeze(np.argwhere(np.logical_and(self.y[:,0] >= 0, self.y[:,0] < newshape[1])))
                colinds = np.squeeze(np.argwhere(np.logical_and(self.x[rowinds,:][0,:] >= 0, self.x[rowinds,:][0,:] < newshape[0])))
                self.crop_inds = (rowinds,colinds)


            elif self.x.ndim == 2:
                xinds = np.argwhere(np.logical_and(self.x >= 0, self.x < newshape[0]))
                yinds = np.argwhere(np.logical_and(self.y[xinds] >= 0, self.y[xinds] < newshape[1]))
                self.crop_inds = (xinds,yinds)
            else:
                raise Exception("Cropping a raydata object with strange dimensions! I don't know how to do that; help!")

            self.crop = copy.deepcopy(window)

        else:
            raise ValueError('Cannot understand detector window; should be None or (Left,Top,Width,Height)')


    def get_ray_start(self,x=None,y=None,im_position_tol = 1,coords='Display'):
        '''
        Get the 3D x,y,z coordinates of the "start" of the casted rays / sightlines.

        Parameters:

            x,y (array-like)        : Image pixel coordinates at which to get the sight-line start coordinates.\
                                      If not specified, the start coordinates of all casted sight lines will be returned.
            im_position_tol (float) : If x and y are specified but no sight-line was cast at exactly the \
                                      input coordinates, the nearest casted sight-line will be returned \
                                      instead provided the pixel coordinates wre within this many pixels of \
                                      the requested coordinates.
            coords (str)            : Either ``Display`` or ``Coords``, specifies what orientation the input x \
                                      and y correspond to or orientation of the returned array.

        Returns:

            np.ndarray              : An array containing the 3D coordinates, in metres, of the start of \
                                      each sight line (i.e. camera pupil position). If x and y coordinates were \
                                      given either to this function or to calcam.raycast_sightlines(), the shape of \
                                      this array is the same as the input x and y arrays with an additional \
                                      dimension added which contains the  [X,Y,Z] 3D coordinates. Otherwise the shape\
                                      is (h x w x 3) where w and h are the image width and height (in display coords).
        '''
        if x is None and y is None:
            if self.fullchip:
                if coords.lower() == 'display':
                    if self.crop is None:
                        return self.ray_start_coords
                    else:
                        return self.ray_start_coords[self.crop_inds[0],:,:][:,self.crop_inds[1],:]
                else:
                    if self.crop is None:
                        return self.transform.display_to_original_image(self.ray_start_coords)
                    else:
                        return self.transform.display_to_original_image(self.ray_start_coords[self.crop_inds[0],:,:][:,self.crop_inds[1],:])
            else:
                if self.crop is None:
                    return self.ray_start_coords
                else:
                    return self.ray_start_coords[self.crop_inds[0],:][self.crop_inds[1],:]
        else:
            if self.x is None or self.y is None:
                raise Exception('This ray data does not have x and y pixel indices!')
            if np.shape(x) != np.shape(y):
                raise ValueError('x and y arrays must be the same shape!')
            else:

                if coords.lower() == 'original':
                    x,y = self.transform.original_to_display_coords(x,y)

                oldshape = np.shape(x)
                x = np.reshape(x,np.size(x),order='F')
                y = np.reshape(y,np.size(y),order='F')
                [start_X,start_Y,start_Z] = np.split(self.ray_start_coords,3,-1)
                start_X = start_X.flatten()
                start_Y = start_Y.flatten()
                start_Z = start_Z.flatten()
                xflat = self.x.flatten()
                yflat = self.y.flatten()
                Xout = np.zeros(np.shape(x))
                Yout = np.zeros(np.shape(x))
                Zout = np.zeros(np.shape(x))
                for pointno in range(x.size):
                    deltaX = xflat - x[pointno]
                    deltaY = yflat - y[pointno]
                    deltaR = np.sqrt(deltaX**2 + deltaY**2)
                    if np.min(deltaR) <= im_position_tol:
                        Xout[pointno] = start_X[np.argmin(deltaR)]
                        Yout[pointno] = start_Y[np.argmin(deltaR)]
                        Zout[pointno] = start_Z[np.argmin(deltaR)]
                    else:
                        raise Exception('No ray-traced pixel within im_position_tol of requested pixel ({:.1f},{:.1f})!'.format(x[pointno],y[pointno]))
                out = np.hstack([Xout,Yout,Zout])

                return np.reshape(out,oldshape + (3,),order='F')


    def get_ray_end(self,x=None,y=None,im_position_tol = 1,coords='Display'):
        '''
        Get the 3D x,y,z coordinates where the casted rays / sightlines intersect the CAD model.

        Parameters:

            x,y (array-like)        : Image pixel coordinates at which to get the sight-line end coordinates.\
                                      If not specified, the end coordinates of all casted sight lines will be returned.
            im_position_tol (float) : If x and y are specified but no sight-line was cast at exactly the \
                                      input coordinates, the nearest casted sight-line will be returned \
                                      instead provided the pixel coordinates wre within this many pixels of \
                                      the requested coordinates.
            coords (str)            : Either ``Display`` or ``Coords``, specifies what orientation the input x \
                                      and y correspond to or orientation of the returned array.

        Returns:

            np.ndarray              : An array containing the 3D coordinates, in metres, of the points where\
                                      each sight line intersects the CAD model. If x and y coordinates were \
                                      given either to this function or to calcam.raycast_sightlines(), the shape of \
                                      this array is the same as the input x and y arrays with an additional \
                                      dimension added which contains the  [X,Y,Z] 3D coordinates. Otherwise the shape\
                                      is (h x w x 3) where w and h are the image width and height (in display coords).

        '''
        if x is None and y is None:
            if self.fullchip:
                if coords.lower() == 'display':
                    if self.crop is None:
                        return self.ray_end_coords
                    else:
                        return self.ray_end_coords[self.crop_inds[0],:,:][:,self.crop_inds[1],:]
                else:
                    if self.crop is None:
                        return self.transform.display_to_original_image(self.ray_end_coords)
                    else:
                        return self.transform.display_to_original_image(self.ray_end_coords[self.crop_inds[0],:,:][:, self.crop_inds[1],:])
            else:
                if self.crop is None:
                    return self.ray_end_coords
                else:
                    return self.ray_end_coords[self.crop_inds[0],:][self.crop_inds[1],:]
        else:
            if self.x is None or self.y is None:
                raise Exception('This ray data does not have x and y pixel indices!')
            if np.shape(x) != np.shape(y):
                raise ValueError('x and y arrays must be the same shape!')
            else:

                if coords.lower() == 'original':
                    x,y = self.transform.original_to_display_coords(x,y)

                oldshape = np.shape(x)
                x = np.reshape(x,np.size(x),order='F')
                y = np.reshape(y,np.size(y),order='F')
                [end_X,end_Y,end_Z] = np.split(self.ray_end_coords,3,-1)
                end_X = end_X.flatten()
                end_Y = end_Y.flatten()
                end_Z = end_Z.flatten()
                xflat = self.x.flatten()
                yflat = self.y.flatten()
                Xout = np.zeros(np.shape(x))
                Yout = np.zeros(np.shape(x))
                Zout = np.zeros(np.shape(x))
                for pointno in range(x.size):
                    deltaX = xflat - x[pointno]
                    deltaY = yflat - y[pointno]
                    deltaR = np.sqrt(deltaX**2 + deltaY**2)
                    if np.min(deltaR) <= im_position_tol:
                        Xout[pointno] = end_X[np.argmin(deltaR)]
                        Yout[pointno] = end_Y[np.argmin(deltaR)]
                        Zout[pointno] = end_Z[np.argmin(deltaR)]
                    else:
                        raise Exception('No ray-traced pixel within im_position_tol of requested pixel!')
                out = np.hstack([Xout,Yout,Zout])

                return np.reshape(out,oldshape + (3,),order='F')


    def get_model_normals(self,x=None,y=None,im_position_tol = 1,coords='Display'):
        '''
        Get the 3D unit normal vectors of the CAD model surface where the camera sight-lines hit the model.
        Only available if calc_normals = True was given when running raycast_sightlines().

        Parameters:

            x,y (array-like)        : Image pixel coordinates at which to get the model normals.\
                                      If not specified, the end coordinates of all casted sight lines will be returned.
            im_position_tol (float) : If x and y are specified but no sight-line was cast at exactly the \
                                      input coordinates, the nearest casted sight-line will be returned \
                                      instead provided the pixel coordinates wre within this many pixels of \
                                      the requested coordinates.
            coords (str)            : Either ``Display`` or ``Coords``, specifies what orientation the input x \
                                      and y correspond to or orientation of the returned array.

        Returns:

            np.ndarray              : An array containing the normal vectors of the CAD model surface where\
                                      each sight line intersects the CAD model. If x and y coordinates were \
                                      given either to this function or to calcam.raycast_sightlines(), the shape of \
                                      this array is the same as the input x and y arrays with an additional \
                                      dimension added which contains the  [X,Y,Z] components of the normals. Otherwise the shape\
                                      is (h x w x 3) where w and h are the image width and height (in display coords).

        '''

        if self.model_normals is None:
            return Exception('Model normals were not calculated when doing the ray-cast. To use this function you must use raycast_sightlines() with calc_normals=True.')

        if x is None and y is None:
            if self.fullchip:
                if coords.lower() == 'display':
                    if self.crop is None:
                        return self.model_normals
                    else:
                        return self.model_normals[self.crop_inds[0],:,:][:,self.crop_inds[1],:]
                else:
                    if self.crop is None:
                        return self.transform.display_to_original_image(self.model_normals)
                    else:
                        return self.transform.display_to_original_image(self.model_normals[self.crop_inds[0],:,:][:, self.crop_inds[1],:])
            else:
                if self.crop is None:
                    return self.model_normals
                else:
                    return self.model_normals[self.crop_inds[0],:][self.crop_inds[1],:]
        else:
            if self.x is None or self.y is None:
                raise Exception('This ray data does not have x and y pixel indices!')
            if np.shape(x) != np.shape(y):
                raise ValueError('x and y arrays must be the same shape!')
            else:

                if coords.lower() == 'original':
                    x,y = self.transform.original_to_display_coords(x,y)

                oldshape = np.shape(x)
                x = np.reshape(x,np.size(x),order='F')
                y = np.reshape(y,np.size(y),order='F')
                [norm_X,norm_Y,norm_Z] = np.split(self.model_normals,3,-1)
                norm_X = norm_X.flatten()
                norm_Y = norm_Y.flatten()
                norm_Z = norm_Z.flatten()
                xflat = self.x.flatten()
                yflat = self.y.flatten()
                Xout = np.zeros(np.shape(x))
                Yout = np.zeros(np.shape(x))
                Zout = np.zeros(np.shape(x))
                for pointno in range(x.size):
                    deltaX = xflat - x[pointno]
                    deltaY = yflat - y[pointno]
                    deltaR = np.sqrt(deltaX**2 + deltaY**2)
                    if np.min(deltaR) <= im_position_tol:
                        Xout[pointno] = norm_X[np.argmin(deltaR)]
                        Yout[pointno] = norm_Y[np.argmin(deltaR)]
                        Zout[pointno] = norm_Z[np.argmin(deltaR)]
                    else:
                        raise Exception('No ray-traced pixel within im_position_tol of requested pixel!')
                out = np.hstack([Xout,Yout,Zout])

                return np.reshape(out,oldshape + (3,),order='F')

    def get_ray_lengths(self,x=None,y=None,im_position_tol = 1,coords='Display'):
        '''
        Get the sight-line lengths either of all casted sight-lines or at the specified image coordinates.

        Parameters:
        
            x,y (array-like)        : Image pixel coordinates at which to get the sight-line lengths. \
                                      If not specified, the lengths of all casted sight lines will be returned.
            im_position_tol (float) : If x and y are specified but no sight-line was cast at exactly the \
                                      input coordinates, the nearest casted sight-line will be returned \
                                      instead provided the pixel coordinates wre within this many pixels of \
                                      the requested coordinates.
            coords (str)            : Either ``Display`` or ``Coords``, specifies what orientation the input x \
                                      and y correspond to or orientation of the returned array.
        
        Returns:

            np.ndarray              : Array containing the sight-line lengths. If the ray cast was for the \
                                      full detector and x and y are not specified, the array shape will be \
                                      (h x w) where w nd h are the image width and height. Otherwise it will \
                                      be the same shape as the input x and y coordinates.
        '''
        
        # Work out ray lengths for all raytraced pixels
        raylength = np.sqrt(np.sum( (self.ray_end_coords - self.ray_start_coords) **2,axis=-1))
        
        # If no x and y given, return them all
        if x is None and y is None:
            if self.fullchip:
                if coords.lower() == 'display':
                    if self.crop is None:
                        return raylength
                    else:
                        return raylength[self.crop_inds[0],:][:,self.crop_inds[1]]
                else:
                    if self.crop is None:
                        return self.transform.display_to_original_image(raylength)
                    else:
                        return self.transform.display_to_original_image(raylength[self.crop_inds[0],:][:,self.crop_inds[1]])
            else:
                if self.crop is None:
                    return raylength
                else:
                    return raylength[self.crop_inds[0]][self.crop_inds[1]]
        else:
            if self.x is None or self.y is None:
                raise Exception('This ray data does not have x and y pixel indices!')

            # Otherwise, return the ones at given x and y pixel coords.
            if np.shape(x) != np.shape(y):
                raise ValueError('x and y arrays must be the same shape!')
            else:

                if coords.lower() == 'original':
                    x,y = self.transform.original_to_display_coords(x,y)

                orig_shape = np.shape(x)
                x = np.reshape(x,np.size(x),order='F')
                y = np.reshape(y,np.size(y),order='F')
                RL = np.zeros(np.shape(x))

                if not self.fullchip:
                    raylength = raylength.flatten()
                    xflat = self.x.flatten()
                    yflat = self.y.flatten()

                for pointno in range(x.size):
                    if np.isnan(x[pointno]) or np.isnan(y[pointno]):
                        RL[pointno] = np.nan
                        continue

                    if self.fullchip:
                        xind = np.argmin(np.abs(self.x[0,:] - x[pointno]))
                        yind = np.argmin(np.abs(self.y[:,0] - y[pointno]))
                        deltaR = np.sqrt( (self.x[0,xind]-x[pointno])**2 + (self.y[yind,0]-y[pointno])**2)
                        if deltaR < im_position_tol:
                            RL[pointno] = raylength[yind,xind]
                        else:
                            raise Exception('No ray-traced pixel within im_position_tol of requested pixel ({:.1f},{:.1f})!'.format(x[pointno], y[pointno]))
                    else:
                        # This can be very slow if xflat and yflat are big arrays!
                        deltaR = np.sqrt( (xflat - x[pointno])**2 + (yflat - y[pointno])**2 )
                        if np.nanmin(deltaR) <= im_position_tol:
                            RL[pointno] = raylength[np.nanargmin(deltaR)]
                        else:
                            raise Exception('No ray-traced pixel within im_position_tol of requested pixel!')

                return np.reshape(RL,orig_shape,order='F')



    def get_ray_directions(self,x=None,y=None,im_position_tol=1,coords='Display'):
        '''
        Get unit vectors specifying the sight-line directions. Note that ray casting \
        is not required to get this information: see :func:`calcam.Calibration.get_los_direction` \
        for the same functionality, however this can be useful if you have the RayData \
        but not :class:`calcam.Calibration` object loaded when doing the analysis.

        Parameters:

            x,y (array-like)        : x and y pixel coordinates at which to get the ray directions. \
                                      If not specified, the ray directions of every sight-line are returned.
            im_position_tol (float) : If x and y are specified but no sight-line was cast at exactly the \
                                      input coordinates, the nearest casted sight-line will be returned \
                                      instead provided the pixel coordinates wre within this many pixels of \
                                      the requested coordinates.
            coords (str)            : Either ``Display`` or ``Original``, specifies what orientation the input x \
                                      and y correspond to or orientation of the returned array.

        Returns:

            np.ndarray              : Array containing the sight-line directions. If the ray cast was for the \
                                      full detector and x and y are not specified, the array shape will be \
                                      (h x w x 3) where w nd h are the image width and height. Otherwise it will \
                                      be the same shape as the input x and y coordinates plus an extra dimension.
        '''
        vectors = (self.ray_end_coords - self.ray_start_coords)
        lengths = np.sqrt(np.sum(vectors**2,axis=-1))
        dirs =  vectors / np.repeat(lengths.reshape(np.shape(lengths)+(1,)),3,axis=-1)

        if x is None and y is None:
            if self.fullchip:
                if coords.lower() == 'display':
                    if self.crop is None:
                        return dirs
                    else:
                        return dirs[self.crop_inds[0],:,:][:,self.crop_inds[1],:]
                else:
                    if self.crop is None:
                        return self.transform.display_to_original_image(dirs)
                    else:
                        return self.transform.display_to_original_image(dirs[self.crop_inds[0],:,:][:, self.crop_inds[1],:])
            else:
                if self.crop is None:
                    return dirs
                else:
                    return dirs[self.crop_inds[0]][self.crop_inds[1]]
        else:
            if self.x is None or self.y is None:
                raise Exception('This ray data does not have x and y pixel indices!')
            if np.shape(x) != np.shape(y):
                raise ValueError('x and y arrays must be the same shape!')
            else:

                if coords.lower() == 'original':
                    x,y = self.transform.original_to_display_coords(x,y)

                oldshape = np.shape(x)
                x = np.reshape(x,np.size(x),order='F')
                y = np.reshape(y,np.size(y),order='F')
                [dirs_X,dirs_Y,dirs_Z] = np.split(dirs,3,-1)
                dirs_X = dirs_X.flatten()
                dirs_Y = dirs_Y.flatten()
                dirs_Z = dirs_Z.flatten()
                xflat = self.x.flatten()
                yflat = self.y.flatten()
                Xout = np.zeros(np.shape(x))
                Yout = np.zeros(np.shape(x))
                Zout = np.zeros(np.shape(x))
                for pointno in range(x.size):
                    deltaX = xflat - x[pointno]
                    deltaY = yflat - y[pointno]
                    deltaR = np.sqrt(deltaX**2 + deltaY**2)
                    if np.min(deltaR) <= im_position_tol:
                        Xout[pointno] = dirs_X[np.argmin(deltaR)]
                        Yout[pointno] = dirs_Y[np.argmin(deltaR)]
                        Zout[pointno] = dirs_Z[np.argmin(deltaR)]
                    else:
                        raise Exception('No ray-traced pixel within im_position_tol of requested pixel!')
                out = np.hstack([Xout,Yout,Zout])

                return np.reshape(out,oldshape + (3,),order='F')



'''
* Copyright 2015-2022 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

"""
CalCam package.
"""
import os
import warnings

# Calcam version
with open(os.path.join(os.path.split(os.path.abspath(__file__))[0],'__version__'),'r') as ver_file:
    __version__ = ver_file.readline().rstrip()

# Calcam supports a "headless" mode which does not require VTK / PyQt. This way it can still be used for analysis
# using calibration or raydata objects even if the GUI libraries are not available.
no_gui_reason = None

try:
    import vtk
    vtk.vtkObject.GlobalWarningDisplayOff()

    try:
        from . import gui
        from .gui import start_gui

    except Exception as e:
        warnings.warn('Error importing GUI modules (error: {:}) - the calcam.gui module will not be available.'.format(e),ImportWarning)
        no_gui_reason = str(e)

    from .cadmodel import CADModel
    from .raycast import raycast_sightlines
    from .render import render_cam_view,render_unfolded_wall

except Exception as e:
    warnings.warn('Cannot import VTK python package (error: {:}) - the calcam.gui, calcam.raycast, calcam.render and calcam.cadmodel modules will not be available.'.format(e),ImportWarning)
    no_gui_reason = 'Cannot import VTK python package (error: {:})'.format(e)

# Import the top level "public facing" classes & functions
# which do not rely on VTK.
from .calibration import Calibration
from . import movement
from .raycast import RayData
from .pointpairs import PointPairs

from . import config
from . import gm

# If we have no GUI available, put a placeholder function to print a message about why where the GUI launcher would normally be.
if no_gui_reason is not None:
    start_gui = lambda : print('Could not start calcam GUI: {:s}'.format(no_gui_reason))



'''
* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

from .core import *
from .vtkinteractorstyles import CalcamInteractorStyle3D
from ..calibration import Calibration
from ..image_enhancement import enhance_image, scale_to_8bit

# View designer window.
# This allows creation of FitResults objects for a 'virtual' camera.
class AlignmentCalib(CalcamGUIWindow):

    def __init__(self, app, parent = None, load_file=None):

        # GUI initialisation
        CalcamGUIWindow.init(self,'alignment_calib_editor.ui',app,parent)

        # Start up with no CAD model
        self.cadmodel = None

        # Set up VTK
        self.qvtkwidget_3d = qt.QVTKRenderWindowInteractor(self.vtk_frame)
        self.vtk_frame.layout().addWidget(self.qvtkwidget_3d,0,0)
        self.interactor3d = CalcamInteractorStyle3D(refresh_callback=self.refresh_3d,viewport_callback=self.update_viewport_info,resize_callback=self.update_vtk_size)
        self.qvtkwidget_3d.SetInteractorStyle(self.interactor3d)
        self.renderer_3d = vtk.vtkRenderer()
        self.renderer_3d.SetBackground(0, 0, 0)
        self.qvtkwidget_3d.GetRenderWindow().AddRenderer(self.renderer_3d)
        self.camera_3d = self.renderer_3d.GetActiveCamera()


        self.populate_models()


        # Synthetic camera object to store the results
        self.calibration = Calibration(cal_type='alignment')
        self.chessboard_fit = None

        self.filename = None
        self.original_image = None

        self.manual_exc = True

        # Callbacks for GUI elements
        self.image_sources_list.currentIndexChanged.connect(self.build_imload_gui)
        self.viewlist.itemSelectionChanged.connect(self.change_cad_view)
        self.camX.valueChanged.connect(self.change_cad_view)
        self.camY.valueChanged.connect(self.change_cad_view)
        self.camZ.valueChanged.connect(self.change_cad_view)
        self.tarX.valueChanged.connect(self.change_cad_view)
        self.tarY.valueChanged.connect(self.change_cad_view)
        self.tarZ.valueChanged.connect(self.change_cad_view)
        self.cam_roll.valueChanged.connect(self.change_cad_view)
        self.load_model_button.clicked.connect(self._load_model)
        self.model_name.currentIndexChanged.connect(self.populate_model_variants)
        self.feature_tree.itemChanged.connect(self.update_checked_features)
        self.feature_tree.itemSelectionChanged.connect(self.update_cadtree_selection)
        self.load_image_button.clicked.connect(self.load_image)
        self.im_flipud.clicked.connect(self.transform_image)
        self.rendertype_edges.toggled.connect(self.toggle_wireframe)
        self.im_fliplr.clicked.connect(self.transform_image)
        self.im_rotate_button.clicked.connect(self.transform_image)
        self.im_reset.clicked.connect(self.transform_image)
        self.im_y_stretch_button.clicked.connect(self.transform_image)
        self.calcam_intrinsics.clicked.connect(self.update_intrinsics)
        self.chessboard_intrinsics.clicked.connect(self.update_intrinsics)
        self.pinhole_intrinsics.clicked.connect(self.update_intrinsics)
        self.pixel_size_checkbox.toggled.connect(self.update_intrinsics)
        self.pixel_size_box.valueChanged.connect(self.update_intrinsics)
        self.focal_length_box.valueChanged.connect(self.update_intrinsics)
        self.load_chessboard_button.clicked.connect(self.update_chessboard_intrinsics)
        self.load_intrinsics_button.clicked.connect(self.update_intrinsics)
        self.load_extrinsics_button.clicked.connect(self.load_viewport_calib)
        self.cad_colour_reset_button.clicked.connect(self.set_cad_colour)
        self.cad_colour_choose_button.clicked.connect(self.set_cad_colour)
        self.im_opacity_slider.valueChanged.connect(self.update_overlay)
        self.enhance.clicked.connect(self.update_overlay)
        self.edge_detect.clicked.connect(self.update_overlay)
        self.no_effect.clicked.connect(self.update_overlay)
        self.im_edge_colour_button.clicked.connect(self.update_edge_colour)
        self.edge_threshold_1.valueChanged.connect(self.update_overlay)
        self.edge_threshold_2.valueChanged.connect(self.update_overlay)
        self.image_mask_button.clicked.connect(self.edit_masking)

        self.control_sensitivity_slider.valueChanged.connect(lambda x: self.interactor3d.set_control_sensitivity(x*0.01))
        self.rmb_rotate.toggled.connect(self.interactor3d.set_rmb_rotate)
        self.interactor3d.set_control_sensitivity(self.control_sensitivity_slider.value()*0.01)

        self.pixel_size_box.setSuffix(u' \u00B5m')

        self.action_save.triggered.connect(self.save)
        self.action_save_as.triggered.connect(lambda: self.save(saveas=True))
        self.action_open.triggered.connect(self.open_calib)
        self.action_new.triggered.connect(self.reset)
        self.action_cal_info.triggered.connect(self.show_calib_info)

        self.viewport_calibs = DodgyDict()

        self.image_settings.hide()
        self.image_display_settings.hide()
        self.edge_detect_settings.hide()

        self.tabWidget.setTabEnabled(2,False)

        # Populate image sources list and tweak GUI layout for image loading.
        self.imload_inputs = {}
        self.image_load_options.layout().setColumnMinimumWidth(0,100)

        self.image_sources = self.config.get_image_sources()
        index = -1
        for i,imsource in enumerate(self.image_sources):
            self.image_sources_list.addItem(imsource.display_name)
            if imsource.display_name == self.config.default_image_source:
                index = i

        self.image_sources_list.setCurrentIndex(index)

        self.edge_detect_colour = (1,0,0)
        self.view_aspect = None
        self.vtk_aspect = None
        self.intrinsics_calib = None

        self.extrinsics_src = None

        self.calibration = None


        # If we have an old version of openCV, histo equilisation won't work :(
        cv2_version = float('.'.join(cv2.__version__.split('.')[:2]))
        cv2_micro_version = int(cv2.__version__.split('.')[2].split('-')[0])
        if cv2_version < 2.4 or (cv2_version == 2.4 and cv2_micro_version < 6):
            self.edge_detect.setEnabled(False)
            self.edge_detect.setToolTip('Requires OpenCV 2.4.6 or newer; you have {:s}'.format(cv2.__version__))

        self.fov_enabled = False
        self.viewdir_at_cc = True

        # Start the GUI!
        self.show()

        self.interactor3d.init()
        self.qvtkwidget_3d.GetRenderWindow().GetInteractor().Initialize()
        self.reset()

        if load_file is not None:
            self.open_calib(load_file)



    def _load_model(self):
        self.load_model(hold_view = self.cadmodel is not None or self.calibration.filename is not None)
        if self.calibration.image is not None:
            self.tabWidget.setTabEnabled(2,True)
        self.update_intrinsics()

    def on_close(self):
        self.qvtkwidget_3d.close()

    def reset(self,keep_cadmodel=False):

        if not keep_cadmodel:
            if self.cadmodel is not None:
                self.cadmodel.remove_from_renderer(self.renderer_3d)
                self.cadmodel.unload()
                self.feature_tree.blockSignals(True)
                self.feature_tree.clear()
                self.feature_tree.blockSignals(False)
                self.cadmodel = None

        self.pinhole_intrinsics.setChecked(True)
        self.pixel_size_checkbox.setChecked(False)

        self.interactor3d.set_overlay_image(None)

        self.calibration = Calibration(cal_type='alignment')

        self.tabWidget.setTabEnabled(2,False)

        self.filename = None
        self.setWindowTitle('Calcam Calibration Tool (Manual Alignment)')

        self.chessboard_fit = None
        self.intrinsics_calib = None

        self.update_intrinsics()

        self.refresh_3d()
        self.unsaved_changes = False


    def edit_masking(self):

        dialog = ImageMaskDialog(self,self.calibration.get_image(),allow_subviews=False)
        result = dialog.exec()

        if result == 1:
            self.calibration.set_subview_mask(dialog.fieldmask,subview_names=dialog.field_names,coords='Display')
            self.unsaved_changes = True

        del dialog

    def update_edge_colour(self):

        col = self.pick_colour(self.edge_detect_colour)

        if col is not None:
            self.edge_detect_colour = col
            self.update_overlay()

    def toggle_wireframe(self,wireframe):

        if self.cadmodel is not None:

            self.cadmodel.set_wireframe( wireframe )

            self.refresh_3d()


    def update_viewport_info(self,keep_selection=False):

        CalcamGUIWindow.update_viewport_info(self,keep_selection)
        self.unsaved_changes = True

        if self.original_image is not None:

            if self.pinhole_intrinsics.isChecked():

                fov = 3.14159 * self.camera_3d.GetViewAngle() / 180.
                f = self.calibration.geometry.get_display_shape()[1]/(2*np.tan(fov/2.))
                if self.pixel_size_checkbox.isChecked():
                        f = f * self.pixel_size_box.value() / 1e3

                self.focal_length_box.blockSignals(True)
                self.focal_length_box.setValue(f)
                self.focal_length_box.blockSignals(False)


    def open_calib(self,filename=None):

        try:
            opened_calib = Calibration(filename)
        except:
            opened_calib = self.object_from_file('calibration')

        if opened_calib is None:
            return

        if opened_calib._type == 'fit':
            raise UserWarning('The selected calibration is a point-pair fitting calibration and cannot be edited in this tool. Please open it with the point fitting calibration tool instead.')
        elif opened_calib._type == 'virtual':
            raise UserWarning('The selected calibration is a virtual calibration and cannot be edited in this tool. Please open it in the virtual calibration editor instead.')

        if opened_calib.cad_config is not None:
            cconfig = opened_calib.cad_config
            if self.cadmodel is not None and self.cadmodel.machine_name == cconfig['model_name'] and self.cadmodel.model_variant == cconfig['model_variant']:
                keep_model = True
            else:
                keep_model = False
        else:
            keep_model = False

        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.reset(keep_cadmodel = keep_model)

        # Basic setup
        self.filename = opened_calib.filename
        self.setWindowTitle('Calcam Calibration Tool (Manual Alignment) - {:s}'.format(os.path.split(self.filename)[-1][:-4]))

        # Load the image
        for imsource in self.image_sources:
            if imsource.display_name == 'Calcam Calibration':
                try:
                    self.load_image(newim = imsource.get_image_function(self.filename))
                except Exception as e:
                    self.show_msgbox(e)

        # Load the appropriate CAD model, if we know what that is
        if opened_calib.cad_config is not None:
            if keep_model:
                try:
                    self.cadmodel.enable_only(cconfig['enabled_features'])
                except Exception as e:
                    if 'Unknown feature' not in str(e):
                        raise
                self.update_feature_tree_checks()
            else:
                cconfig = opened_calib.cad_config
                load_model = True
                try:
                    name_index = sorted(self.model_list.keys()).index(cconfig['model_name'])
                    self.model_name.setCurrentIndex(name_index)
                    variant_index = self.model_list[ cconfig['model_name'] ][1].index(cconfig['model_variant'])
                    self.model_variant.setCurrentIndex(variant_index)
                except ValueError:
                    self.model_name.setCurrentIndex(-1)
                    load_model=False


                if load_model:
                    try:
                        self.load_model(featurelist=cconfig['enabled_features'],hold_view=True)
                    except Exception as e:
                        self.cadmodel = None
                        if 'Unknown feature' not in str(e):
                            raise


        self.calibration = opened_calib

        if self.calibration.readonly:
            self.action_save.setEnabled(False)
        else:
            self.action_save.setEnabled(True)

        # Load the intrinsics
        if opened_calib.intrinsics_type == 'pinhole':
            fl = opened_calib.view_models[0].cam_matrix[0,0]
            if opened_calib.pixel_size is not None:
                fl = fl * opened_calib.pixel_size  * 1000

                self.pixel_size_box.setValue(opened_calib.pixel_size*1e6)
                self.pixel_size_checkbox.setChecked(True)
            else:
                self.pixel_size_checkbox.setChecked(False)

            self.focal_length_box.setValue(fl)
            self.pinhole_intrinsics.setChecked(True)

        elif opened_calib.intrinsics_type == 'chessboard':
            self.chessboard_fit = opened_calib.view_models[0]
            self.chessboard_pointpairs = opened_calib.intrinsics_constraints
            self.chessboard_src = opened_calib.history['intrinsics']
            self.chessboard_intrinsics.setChecked(True)

        elif opened_calib.intrinsics_type == 'calibration':
            self.intrinsics_calib = self.calibration
            self.calcam_intrinsics.setChecked(True)

        self.update_intrinsics()

        self.set_view_from_calib(self.calibration,0)

        self.app.restoreOverrideCursor()
        if self.cadmodel is not None and self.calibration.image is not None:
            self.tabWidget.setTabEnabled(2,True)

        self.unsaved_changes = False


    def update_intrinsics(self):

        if self.calibration.image is None:
            return

        nx,ny = self.calibration.geometry.get_original_shape()
        self.calibration.set_subview_mask(np.zeros((ny,nx),dtype=np.int8), coords='Original')

        if self.sender() is self.load_intrinsics_button:
            self.intrinsics_calib = None

        if self.pixel_size_checkbox.isChecked():
            self.pixel_size_box.setEnabled(True)
            self.calibration.pixel_size = self.pixel_size_box.value() / 1e6
        else:
            self.pixel_size_box.setEnabled(False)
            self.calibration.pixel_size = None

        if self.calcam_intrinsics.isChecked():
            self.interactor3d.zoom_enabled = False
            self.load_chessboard_button.setEnabled(False)
            self.load_intrinsics_button.setEnabled(True)
            self.focal_length_box.setEnabled(False)

            if self.intrinsics_calib is None:
                self.intrinsics_calib = self.object_from_file('calibration')

            if self.intrinsics_calib is not None:
                if len(self.intrinsics_calib.view_models) != 1:
                    self.intrinsics_calib = None
                    self.current_intrinsics_combobox.setChecked(True)
                    raise UserWarning('This calibration has multiple sub-fields, which is not supported by the manual alignment calibration tool.')

                self.calibration.set_calib_intrinsics(self.intrinsics_calib,update_hist_recursion = not (self.intrinsics_calib is self.calibration))
                self.calibration.set_subview_mask(self.intrinsics_calib.get_subview_mask(coords='Original'),coords='Original')
                self.current_intrinsics_combobox = self.calcam_intrinsics
            else:
                self.current_intrinsics_combobox.setChecked(True)


        elif self.pinhole_intrinsics.isChecked():
            self.interactor3d.zoom_enabled = True

            if self.pixel_size_checkbox.isChecked():
                if self.focal_length_box.suffix() == ' px':
                    self.focal_length_box.blockSignals(True)
                    self.focal_length_box.setValue( self.focal_length_box.value() * self.pixel_size_box.value() / 1e3 )
                    self.focal_length_box.blockSignals(False)
                    self.focal_length_box.setSuffix(' mm')
            else:
                if self.focal_length_box.suffix() == ' mm':
                    self.focal_length_box.blockSignals(True)
                    self.focal_length_box.setValue( 1e3 * self.focal_length_box.value() / self.pixel_size_box.value() )
                    self.focal_length_box.blockSignals(False)
                    self.focal_length_box.setSuffix(' px')

            self.load_chessboard_button.setEnabled(False)
            self.focal_length_box.setEnabled(True)

            nx,ny = self.calibration.geometry.get_display_shape()

            f = self.focal_length_box.value()
            if self.pixel_size_checkbox.isChecked():
                f = 1e3 * f / self.pixel_size_box.value()

            self.calibration.set_pinhole_intrinsics(fx=f,fy=f)
            self.current_intrinsics_combobox = self.pinhole_intrinsics

        elif self.chessboard_intrinsics.isChecked():
            self.interactor3d.zoom_enabled = False
            self.load_chessboard_button.setEnabled(True)
            self.load_intrinsics_button.setEnabled(False)
            self.focal_length_box.setEnabled(False)

            if self.chessboard_fit is None:
                self.update_chessboard_intrinsics()
                if self.chessboard_fit is None:
                    self.current_intrinsics_combobox.setChecked(True)
                return
            else:
                self.calibration.set_chessboard_intrinsics(self.chessboard_fit,self.chessboard_pointpairs,self.chessboard_src)
                self.current_intrinsics_combobox = self.chessboard_intrinsics


        self.update_overlay()

        mat = self.calibration.get_cam_matrix()
        n = self.calibration.geometry.get_display_shape()
        wcx = -2.*(mat[0,2] - n[0]/2.) / float(n[0])
        wcy = 2.*(mat[1,2] - n[1]/2.) / float(n[1])
        self.camera_3d.SetWindowCenter(wcx,wcy)
        fov = 360*np.arctan( float(n[1]) / (2*mat[1,1]))/3.14159
        self.camera_3d.SetViewAngle(fov)
        self.unsaved_changes = True



    def on_load_image(self,newim):

        if newim['pixel_size'] is not None:
            self.pixel_size_checkbox.setChecked(True)
            self.pixel_size_box.setValue(newim['pixel_size']*1e6)
        else:
            self.pixel_size_checkbox.setChecked(False)

        self.calibration.set_image( scale_to_8bit(newim['image_data']) , newim['source'],subview_mask = newim['subview_mask'], transform_actions = newim['transform_actions'],coords=newim['coords'],subview_names=newim['subview_names'],pixel_aspect=newim['pixel_aspect'],pixel_size=newim['pixel_size'],offset=newim['image_offset'] )

        imshape = self.calibration.geometry.get_display_shape()
        self.interactor3d.force_aspect = float( imshape[1] ) / float( imshape[0] )

        # This is a slight hack - we have to resize the window slightly
        # and resize it back again to get VTK to redraw the background.
        # More elegant solution needed here really.
        size = self.size()
        self.resize(size.width()+1,size.height())
        self.refresh_3d()
        self.resize(size.width(),size.height())

        self.calibration.view_models = [None] * self.calibration.n_subviews
        self.fit_timestamps = [None] * self.calibration.n_subviews

        if self.calcam_intrinsics.isChecked() and np.any(np.array(self.intrinsics_calib.geometry.get_original_shape()) != np.array(self.calibration.geometry.get_original_shape())):
            self.show_msgbox('The current calibration intrinsics are the wrong shape for this image. The current intrinsics will be reset.')
            self.intrinsics_calib = None
            self.pinhole_intrinsics.setChecked(True)

        self.update_intrinsics()

        self.image_settings.show()
        self.image_display_settings.show()

        if self.cadmodel is not None:
            self.tabWidget.setTabEnabled(2,True)

        self.update_image_info_string(newim['image_data'],self.calibration.geometry)
        self.update_overlay()
        self.unsaved_changes = True


    def update_overlay(self):

        self.overlay_image = self.calibration.undistort_image( self.calibration.get_image(coords='display') )

        if self.enhance.isChecked() or self.edge_detect.isChecked():
            self.overlay_image = enhance_image(self.overlay_image)

        if self.edge_detect.isChecked():

            self.edge_detect_settings.show()
            self.label_18.hide()
            self.im_opacity_slider.hide()
            if self.edge_threshold_1.value() > self.edge_threshold_2.value():
                if self.sender() is self.edge_threshold_1:
                    self.edge_threshold_2.setValue(self.edge_threshold_1.value())
                else:
                    self.edge_threshold_1.setValue(self.edge_threshold_2.value())

            blurred_im = cv2.GaussianBlur(self.overlay_image.astype('uint8'),(5,5),0)
            edge_im = cv2.Canny(blurred_im,self.edge_threshold_1.value(),self.edge_threshold_2.value())
            self.overlay_image = np.zeros(edge_im.shape + (4,),dtype='uint8')
            self.overlay_image[:,:,0] = edge_im[:,:]*self.edge_detect_colour[0]
            self.overlay_image[:,:,1] = edge_im[:,:]*self.edge_detect_colour[1]
            self.overlay_image[:,:,2] = edge_im[:,:]*self.edge_detect_colour[2]
            self.overlay_image[:,:,3] = 255*(edge_im > 0)
        else:
            self.edge_detect_settings.hide()
            self.im_opacity_slider.show()
            self.label_18.show()

        if len(self.overlay_image.shape) < 3:
            self.overlay_image = np.tile(self.overlay_image[:,:,np.newaxis],(1,1,3))

        if self.overlay_image.shape[2] == 3:
            alpha = np.uint8(np.ones(self.overlay_image.shape[:2])*255)
            self.overlay_image = np.dstack([self.overlay_image,alpha])

        if self.im_opacity_slider.isEnabled():
            alpha = self.im_opacity_slider.value()/10.
        else:
            alpha = 1.

        self.interactor3d.set_overlay_image( (self.overlay_image * np.tile([1,1,1,alpha],self.overlay_image.shape[:2] + (1,))).astype('uint8'))

        self.refresh_3d()


    def transform_image(self,data):

        if self.sender() is self.im_flipud:
            self.calibration.geometry.add_transform_action('flip_up_down')

        elif self.sender() is self.im_fliplr:
            self.calibration.geometry.add_transform_action('flip_left_right')

        elif self.sender() is self.im_rotate_button:
            self.calibration.geometry.add_transform_action('rotate_clockwise_{:d}'.format(self.im_rotate_angle.value()))

        elif self.sender() is self.im_y_stretch_button:
            self.calibration.geometry.set_pixel_aspect(self.im_y_stretch_factor.value(),absolute=False)

        elif self.sender() is self.im_reset:
            self.calibration.geometry.set_transform_actions([])
            self.calibration.geometry.set_pixel_aspect(1.)

        # Update the image and point pairs
        imshape = self.calibration.geometry.get_display_shape()
        self.interactor3d.force_aspect = float(imshape[1]) / float(imshape[0])

        self.update_image_info_string(self.calibration.get_image(),self.calibration.geometry)

        self.update_overlay()
        self.unsaved_changes = True





    def save(self,saveas=False):

        if self.calibration.view_models[0] is None:
            raise UserWarning('Nothing to save! You need to load a camera image to calibrate before you can save anything in this tool.')

        if saveas:
            orig_filename = self.filename
            self.filename = None

        if self.filename is None:
            self.filename = self.get_save_filename('calibration')

        if self.filename is not None:

            self.update_extrinsics()

            if self.cadmodel is not None:
                viewport = {'cam_x':self.camX.value(),'cam_y':self.camY.value(),'cam_z':self.camZ.value(),'tar_x':self.tarX.value(),'tar_y':self.tarY.value(),'tar_z':self.tarZ.value(),'fov':self.camera_3d.GetViewAngle(),'roll':self.cam_roll.value()}
                self.calibration.cad_config = {'model_name':self.cadmodel.machine_name , 'model_variant':self.cadmodel.model_variant , 'enabled_features':self.cadmodel.get_enabled_features(),'viewport':viewport }


            self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
            self.statusbar.showMessage('Saving...')
            try:
                self.calibration.save(self.filename)
            except PermissionError:
                raise UserWarning('Could not write to {:s}: permission denied.'.format(self.filename))
            self.unsaved_changes = False
            self.statusbar.clearMessage()
            self.app.restoreOverrideCursor()

            self.action_save.setEnabled(True)
            self.setWindowTitle('Calcam Calibration Tool (Manual Alignment) - {:s}'.format(os.path.split(self.filename)[-1][:-4]))

        elif saveas:
            self.filename = orig_filename
